from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widget import Widget


async def test_visibility_changes() -> None:
    """Test changing visibility via code and CSS.

    See https://github.com/Textualize/textual/issues/1355 as the motivation for these tests.
    """

    class VisibleTester(App[None]):
        """An app for testing visibility changes."""

        CSS = """
        Widget {
            height: 1fr;
        }
        .hidden {
            visibility: hidden;
        }
        """

        def compose(self) -> ComposeResult:
            yield VerticalScroll(
                Widget(id="keep"), Widget(id="hide-via-code"), Widget(id="hide-via-css")
            )

    async with VisibleTester().run_test() as pilot:
        assert pilot.app.query_one("#keep").visible is True
        assert pilot.app.query_one("#hide-via-code").visible is True
        assert pilot.app.query_one("#hide-via-css").visible is True

        pilot.app.query_one("#hide-via-code").styles.visibility = "hidden"
        await pilot.pause(0)
        assert pilot.app.query_one("#keep").visible is True
        assert pilot.app.query_one("#hide-via-code").visible is False
        assert pilot.app.query_one("#hide-via-css").visible is True

        pilot.app.query_one("#hide-via-css").set_class(True, "hidden")
        await pilot.pause(0)
        assert pilot.app.query_one("#keep").visible is True
        assert pilot.app.query_one("#hide-via-code").visible is False
        assert pilot.app.query_one("#hide-via-css").visible is False


async def test_visible_is_inherited() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3071"""

    class InheritedVisibilityApp(App[None]):
        CSS = """
        #four {
            visibility: visible;
        }

        #six {
            visibility: hidden;
        }
        """

        def compose(self):
            yield Widget(id="one")
            with VerticalScroll(id="two"):
                yield Widget(id="three")
                with VerticalScroll(id="four"):
                    yield Widget(id="five")
                    with VerticalScroll(id="six"):
                        yield Widget(id="seven")

    app = InheritedVisibilityApp()
    async with app.run_test():
        assert app.query_one("#one").visible
        assert app.query_one("#two").visible
        assert app.query_one("#three").visible
        assert app.query_one("#four").visible
        assert app.query_one("#five").visible
        assert not app.query_one("#six").visible
        assert not app.query_one("#seven").visible



from decimal import Decimal

import pytest
from rich.style import Style

from textual.color import Color
from textual.css.errors import StyleValueError
from textual.css.scalar import Scalar, Unit
from textual.css.styles import RenderStyles, Styles
from textual.dom import DOMNode
from textual.widget import Widget


def test_styles_reset():
    styles = Styles()
    styles.text_style = "not bold"
    assert styles.text_style == Style(bold=False)
    styles.reset()
    assert styles.text_style is Style.null()


def test_has_rule():
    styles = Styles()
    assert not styles.has_rule("text_style")
    styles.text_style = "bold"
    assert styles.has_rule("text_style")
    styles.text_style = None
    assert not styles.has_rule("text_style")


def test_clear_rule():
    styles = Styles()
    styles.text_style = "bold"
    assert styles.has_rule("text_style")
    styles.clear_rule("text_style")
    assert not styles.has_rule("text_style")


def test_get_rules():
    styles = Styles()
    # Empty rules at start
    assert styles.get_rules() == {}
    styles.text_style = "bold"
    assert styles.get_rules() == {"text_style": Style.parse("bold")}
    styles.display = "none"
    assert styles.get_rules() == {
        "text_style": Style.parse("bold"),
        "display": "none",
    }


def test_set_rule():
    styles = Styles()
    assert styles.get_rules() == {}
    styles.set_rule("text_style", Style.parse("bold"))
    assert styles.get_rules() == {"text_style": Style.parse("bold")}


def test_reset():
    styles = Styles()
    assert styles.get_rules() == {}
    styles.set_rule("text_style", Style.parse("bold"))
    assert styles.get_rules() == {"text_style": Style.parse("bold")}
    styles.reset()
    assert styles.get_rules() == {}


def test_merge():
    styles = Styles()
    styles.set_rule("text_style", Style.parse("bold"))
    styles2 = Styles()
    styles2.set_rule("display", "none")
    styles.merge(styles2)
    assert styles.get_rules() == {
        "text_style": Style.parse("bold"),
        "display": "none",
    }


def test_merge_rules():
    styles = Styles()
    styles.set_rule("text_style", Style.parse("bold"))
    styles.merge_rules({"display": "none"})
    assert styles.get_rules() == {
        "text_style": Style.parse("bold"),
        "display": "none",
    }


def test_render_styles_border():
    base = Styles()
    inline = Styles()
    styles_view = RenderStyles(None, base, inline)

    base.border_top = ("heavy", "red")
    # Base has border-top: heavy red
    assert styles_view.border_top == ("heavy", Color.parse("red"))

    inline.border_left = ("rounded", "green")
    # Base has border-top heavy red, inline has border-left: rounded green
    assert styles_view.border_top == ("heavy", Color.parse("red"))
    assert styles_view.border_left == ("rounded", Color.parse("green"))
    assert styles_view.border == (
        ("heavy", Color.parse("red")),
        ("", Color(0, 255, 0)),
        ("", Color(0, 255, 0)),
        ("rounded", Color.parse("green")),
    )


def test_get_opacity_default():
    styles = RenderStyles(DOMNode(), Styles(), Styles())
    assert styles.text_opacity == 1.0


def test_styles_css_property():
    css = "opacity: 50%; text-opacity: 20%; background: green; color: red; tint: dodgerblue 20%;"
    styles = Styles().parse(css, read_from=("", ""))
    assert styles.css == (
        "background: #008000;\n"
        "color: #FF0000;\n"
        "opacity: 0.5;\n"
        "text-opacity: 0.2;\n"
        "tint: rgba(30,144,255,0.2);"
    )


@pytest.mark.parametrize(
    "set_value, expected",
    [
        [0.2, 0.2],
        [-0.4, 0.0],
        [5.8, 1.0],
        ["25%", 0.25],
        ["-10%", 0.0],
        ["120%", 1.0],
    ],
)
def test_opacity_set_then_get(set_value, expected):
    styles = RenderStyles(DOMNode(), Styles(), Styles())
    styles.text_opacity = set_value
    assert styles.text_opacity == expected


def test_opacity_set_invalid_type_error():
    styles = RenderStyles(DOMNode(), Styles(), Styles())
    with pytest.raises(StyleValueError):
        styles.text_opacity = "invalid value"


def test_opacity_set_allows_integer_value():
    """Regression test for https://github.com/Textualize/textual/issues/3414"""
    styles = RenderStyles(DOMNode(), Styles(), Styles())
    styles.text_opacity = 0
    assert styles.text_opacity == 0.0


@pytest.mark.parametrize(
    "size_dimension_input,size_dimension_expected_output",
    [
        # fmt: off
        [None, None],
        [1, Scalar(1, Unit.CELLS, Unit.WIDTH)],
        [1.0, Scalar(1.0, Unit.CELLS, Unit.WIDTH)],
        [1.2, Scalar(1.2, Unit.CELLS, Unit.WIDTH)],
        [1.2e3, Scalar(1200.0, Unit.CELLS, Unit.WIDTH)],
        ["20", Scalar(20, Unit.CELLS, Unit.WIDTH)],
        ["1.4", Scalar(1.4, Unit.CELLS, Unit.WIDTH)],
        [Scalar(100, Unit.CELLS, Unit.WIDTH), Scalar(100, Unit.CELLS, Unit.WIDTH)],
        [Scalar(10.3, Unit.CELLS, Unit.WIDTH), Scalar(10.3, Unit.CELLS, Unit.WIDTH)],
        [Scalar(10.4, Unit.CELLS, Unit.HEIGHT), Scalar(10.4, Unit.CELLS, Unit.HEIGHT)],
        [Scalar(10.5, Unit.PERCENT, Unit.WIDTH), Scalar(10.5, Unit.WIDTH, Unit.WIDTH)],
        [Scalar(10.6, Unit.PERCENT, Unit.PERCENT), Scalar(10.6, Unit.WIDTH, Unit.WIDTH)],
        [Scalar(10.7, Unit.HEIGHT, Unit.PERCENT), Scalar(10.7, Unit.HEIGHT, Unit.PERCENT)],
        # percentage values are normalised to floats and get the WIDTH "percent_unit":
        [Scalar(11, Unit.PERCENT, Unit.HEIGHT), Scalar(11.0, Unit.WIDTH, Unit.WIDTH)],
        # fmt: on
    ],
)
def test_widget_style_size_can_accept_various_data_types_and_normalize_them(
    size_dimension_input, size_dimension_expected_output
):
    widget = Widget()

    widget.styles.width = size_dimension_input
    assert widget.styles.width == size_dimension_expected_output


@pytest.mark.parametrize(
    "size_dimension_input",
    [
        "a",
        "1.4e3",
        3.14j,
        Decimal("3.14"),
        list(),
        tuple(),
        dict(),
    ],
)
def test_widget_style_size_fails_if_data_type_is_not_supported(size_dimension_input):
    widget = Widget()

    with pytest.raises(StyleValueError):
        widget.styles.width = size_dimension_input



import asyncio

from textual.app import App


class NotificationApp(App[None]):
    pass


async def test_app_no_notifications() -> None:
    """An app with no notifications should have an empty notification list."""
    async with NotificationApp().run_test() as pilot:
        assert len(pilot.app._notifications) == 0


async def test_app_with_notifications() -> None:
    """An app with notifications should have notifications in the list."""
    async with NotificationApp().run_test() as pilot:
        pilot.app.notify("test")
        await pilot.pause()
        assert len(pilot.app._notifications) == 1


async def test_app_with_removing_notifications() -> None:
    """An app with notifications should have notifications in the list, which can be removed."""
    async with NotificationApp().run_test() as pilot:
        pilot.app.notify("test")
        await pilot.pause()
        assert len(pilot.app._notifications) == 1
        pilot.app._unnotify(list(pilot.app._notifications)[0])
        assert len(pilot.app._notifications) == 0


async def test_app_with_notifications_that_expire() -> None:
    """Notifications should expire from an app."""
    async with NotificationApp().run_test() as pilot:
        for n in range(10):
            pilot.app.notify("test", timeout=(0.01 if bool(n % 2) else 60))

        # Wait until the 0.01 timeout expires on all notifications (plus some leeway)
        await asyncio.sleep(0.25)
        assert len(pilot.app._notifications) == 5


async def test_app_clearing_notifications() -> None:
    """The application should be able to clear all notifications."""
    async with NotificationApp().run_test() as pilot:
        for _ in range(100):
            pilot.app.notify("test", timeout=120)
        await pilot.pause()
        assert len(pilot.app._notifications) == 100
        pilot.app.clear_notifications()
        assert len(pilot.app._notifications) == 0



from textual.app import App, ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Button, Label, ProgressBar


class Counter(Widget):
    DEFAULT_CSS = "Counter { height: auto; }"
    counter = reactive(0)  # (1)!

    def compose(self) -> ComposeResult:
        yield Label()
        yield Button("+10")

    def on_button_pressed(self) -> None:
        self.counter += 10

    def watch_counter(self, counter_value: int):
        self.query_one(Label).update(str(counter_value))


class WatchApp(App[None]):
    def compose(self) -> ComposeResult:
        yield Counter()
        yield ProgressBar(total=100, show_eta=False)

    def on_mount(self):
        def update_progress(counter_value: int):  # (2)!
            self.query_one(ProgressBar).update(progress=counter_value)

        self.watch(self.query_one(Counter), "counter", update_progress)  # (3)!


if __name__ == "__main__":
    WatchApp().run()



from textual.app import App, ComposeResult
from textual.reactive import reactive
from textual.widgets import Input, Label


class MultiGreet(App):
    names: reactive[list[str]] = reactive(list, recompose=True)  # (1)!

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Give me a name")
        for name in self.names:
            yield Label(f"Hello, {name}")

    def on_input_submitted(self, event: Input.Changed) -> None:
        self.names.append(event.value)
        self.mutate_reactive(MultiGreet.names)  # (2)!


if __name__ == "__main__":
    app = MultiGreet()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Static

HATCHES = ("cross", "horizontal", "custom", "left", "right")


class HatchApp(App):
    CSS_PATH = "hatch.tcss"

    def compose(self) -> ComposeResult:
        with Horizontal():
            for hatch in HATCHES:
                static = Static(classes=f"hatch {hatch}")
                static.border_title = hatch
                with Vertical():
                    yield static


if __name__ == "__main__":
    app = HatchApp()
    app.run()



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widget import Widget


class NotifyWidget(Widget):
    def on_mount(self) -> None:
        self.notify("test", timeout=60)


class NotifyScreen(Screen):
    def on_mount(self) -> None:
        self.notify("test", timeout=60)

    def compose(self) -> ComposeResult:
        yield NotifyWidget()


class NotifyApp(App[None]):
    def on_mount(self) -> None:
        self.notify("test", timeout=60)
        self.push_screen(NotifyScreen())


async def test_all_levels_of_notification() -> None:
    """All levels within the DOM should be able to notify."""
    async with NotifyApp().run_test() as pilot:
        assert len(pilot.app._notifications) == 3



from string import punctuation

import pytest

from textual import events, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Center, Middle
from textual.css.errors import StylesheetError
from textual.pilot import OutOfBounds
from textual.screen import Screen
from textual.widgets import Button, Label
from textual.worker import WorkerFailed

KEY_CHARACTERS_TO_TEST = "akTW03" + punctuation
"""Test some "simple" characters (letters + digits) and all punctuation."""


class CenteredButtonApp(App):
    CSS = """  # Ensure the button is 16 x 3
    Button {
        min-width: 16;
        max-width: 16;
        width: 16;
        min-height: 3;
        max-height: 3;
        height: 3;
    }
    """

    def compose(self):
        with Center():
            with Middle():
                yield Button()


class ManyLabelsApp(App):
    """Auxiliary app with a button following many labels."""

    AUTO_FOCUS = None  # So that there's no auto-scrolling.

    def compose(self):
        for idx in range(100):
            yield Label(f"label {idx}", id=f"label{idx}")
        yield Button()


async def test_pilot_press_ascii_chars():
    """Test that the pilot can press most ASCII characters as keys."""
    keys_pressed = []

    class TestApp(App[None]):
        def on_key(self, event: events.Key) -> None:
            keys_pressed.append(event.character)

    async with TestApp().run_test() as pilot:
        for char in KEY_CHARACTERS_TO_TEST:
            await pilot.press(char)
            assert keys_pressed[-1] == char


async def test_pilot_exception_catching_app_compose():
    """Ensuring that test frameworks are aware of exceptions
    inside compose methods when running via Pilot run_test()."""

    class FailingApp(App):
        def compose(self) -> ComposeResult:
            1 / 0
            yield Label("Beep")

    with pytest.raises(ZeroDivisionError):
        async with FailingApp().run_test():
            pass


async def test_pilot_exception_catching_widget_compose():
    class SomeScreen(Screen[None]):
        def compose(self) -> ComposeResult:
            1 / 0
            yield Label("Beep")

    class FailingApp(App[None]):
        def on_mount(self) -> None:
            self.push_screen(SomeScreen())

    with pytest.raises(ZeroDivisionError):
        async with FailingApp().run_test():
            pass


async def test_pilot_exception_catching_action():
    """Ensure that exceptions inside action handlers are presented
    to the test framework when running via Pilot run_test()."""

    class FailingApp(App):
        BINDINGS = [Binding("b", "beep", "beep")]

        def action_beep(self) -> None:
            1 / 0

    with pytest.raises(ZeroDivisionError):
        async with FailingApp().run_test() as pilot:
            await pilot.press("b")


async def test_pilot_exception_catching_worker():
    class SimpleAppThatCrashes(App[None]):
        def on_mount(self) -> None:
            self.crash()

        @work(name="crash")
        async def crash(self) -> None:
            1 / 0

    with pytest.raises(WorkerFailed) as exc:
        async with SimpleAppThatCrashes().run_test():
            pass
        assert exc.type is ZeroDivisionError


async def test_pilot_click_screen():
    """Regression test for https://github.com/Textualize/textual/issues/3395.

    Check we can use `Screen` as a selector for a click."""

    async with App().run_test() as pilot:
        await pilot.click("Screen")


async def test_pilot_hover_screen():
    """Regression test for https://github.com/Textualize/textual/issues/3395.

    Check we can use `Screen` as a selector for a hover."""

    async with App().run_test() as pilot:
        await pilot.hover("Screen")


@pytest.mark.parametrize(
    ["method", "screen_size", "offset"],
    [
        #
        ("click", (80, 24), (100, 12)),  # Right of screen.
        ("click", (80, 24), (100, 36)),  # Bottom-right of screen.
        ("click", (80, 24), (50, 36)),  # Under screen.
        ("click", (80, 24), (-10, 36)),  # Bottom-left of screen.
        ("click", (80, 24), (-10, 12)),  # Left of screen.
        ("click", (80, 24), (-10, -2)),  # Top-left of screen.
        ("click", (80, 24), (50, -2)),  # Above screen.
        ("click", (80, 24), (100, -2)),  # Top-right of screen.
        #
        ("click", (5, 5), (7, 3)),  # Right of screen.
        ("click", (5, 5), (7, 7)),  # Bottom-right of screen.
        ("click", (5, 5), (3, 7)),  # Under screen.
        ("click", (5, 5), (-1, 7)),  # Bottom-left of screen.
        ("click", (5, 5), (-1, 3)),  # Left of screen.
        ("click", (5, 5), (-1, -1)),  # Top-left of screen.
        ("click", (5, 5), (3, -1)),  # Above screen.
        ("click", (5, 5), (7, -1)),  # Top-right of screen.
        #
        ("hover", (80, 24), (100, 12)),  # Right of screen.
        ("hover", (80, 24), (100, 36)),  # Bottom-right of screen.
        ("hover", (80, 24), (50, 36)),  # Under screen.
        ("hover", (80, 24), (-10, 36)),  # Bottom-left of screen.
        ("hover", (80, 24), (-10, 12)),  # Left of screen.
        ("hover", (80, 24), (-10, -2)),  # Top-left of screen.
        ("hover", (80, 24), (50, -2)),  # Above screen.
        ("hover", (80, 24), (100, -2)),  # Top-right of screen.
        #
        ("hover", (5, 5), (7, 3)),  # Right of screen.
        ("hover", (5, 5), (7, 7)),  # Bottom-right of screen.
        ("hover", (5, 5), (3, 7)),  # Under screen.
        ("hover", (5, 5), (-1, 7)),  # Bottom-left of screen.
        ("hover", (5, 5), (-1, 3)),  # Left of screen.
        ("hover", (5, 5), (-1, -1)),  # Top-left of screen.
        ("hover", (5, 5), (3, -1)),  # Above screen.
        ("hover", (5, 5), (7, -1)),  # Top-right of screen.
        #
        ("mouse_down", (80, 24), (100, 12)),  # Right of screen.
        ("mouse_down", (80, 24), (100, 36)),  # Bottom-right of screen.
        ("mouse_down", (80, 24), (50, 36)),  # Under screen.
        ("mouse_down", (80, 24), (-10, 36)),  # Bottom-left of screen.
        ("mouse_down", (80, 24), (-10, 12)),  # Left of screen.
        ("mouse_down", (80, 24), (-10, -2)),  # Top-left of screen.
        ("mouse_down", (80, 24), (50, -2)),  # Above screen.
        ("mouse_down", (80, 24), (100, -2)),  # Top-right of screen.
        #
        ("mouse_down", (5, 5), (7, 3)),  # Right of screen.
        ("mouse_down", (5, 5), (7, 7)),  # Bottom-right of screen.
        ("mouse_down", (5, 5), (3, 7)),  # Under screen.
        ("mouse_down", (5, 5), (-1, 7)),  # Bottom-left of screen.
        ("mouse_down", (5, 5), (-1, 3)),  # Left of screen.
        ("mouse_down", (5, 5), (-1, -1)),  # Top-left of screen.
        ("mouse_down", (5, 5), (3, -1)),  # Above screen.
        ("mouse_down", (5, 5), (7, -1)),  # Top-right of screen.
        #
        ("mouse_up", (80, 24), (100, 12)),  # Right of screen.
        ("mouse_up", (80, 24), (100, 36)),  # Bottom-right of screen.
        ("mouse_up", (80, 24), (50, 36)),  # Under screen.
        ("mouse_up", (80, 24), (-10, 36)),  # Bottom-left of screen.
        ("mouse_up", (80, 24), (-10, 12)),  # Left of screen.
        ("mouse_up", (80, 24), (-10, -2)),  # Top-left of screen.
        ("mouse_up", (80, 24), (50, -2)),  # Above screen.
        ("mouse_up", (80, 24), (100, -2)),  # Top-right of screen.
        #
        ("mouse_up", (5, 5), (7, 3)),  # Right of screen.
        ("mouse_up", (5, 5), (7, 7)),  # Bottom-right of screen.
        ("mouse_up", (5, 5), (3, 7)),  # Under screen.
        ("mouse_up", (5, 5), (-1, 7)),  # Bottom-left of screen.
        ("mouse_up", (5, 5), (-1, 3)),  # Left of screen.
        ("mouse_up", (5, 5), (-1, -1)),  # Top-left of screen.
        ("mouse_up", (5, 5), (3, -1)),  # Above screen.
        ("mouse_up", (5, 5), (7, -1)),  # Top-right of screen.
    ],
)
async def test_pilot_target_outside_screen_errors(method, screen_size, offset):
    """Make sure that targeting a click/hover completely outside of the screen errors."""
    app = App()
    async with app.run_test(size=screen_size) as pilot:
        pilot_method = getattr(pilot, method)
        with pytest.raises(OutOfBounds):
            await pilot_method(offset=offset)


@pytest.mark.parametrize(
    ["method", "offset"],
    [
        ("click", (0, 0)),  # Top-left corner.
        ("click", (40, 0)),  # Top edge.
        ("click", (79, 0)),  # Top-right corner.
        ("click", (79, 12)),  # Right edge.
        ("click", (79, 23)),  # Bottom-right corner.
        ("click", (40, 23)),  # Bottom edge.
        ("click", (40, 23)),  # Bottom-left corner.
        ("click", (0, 12)),  # Left edge.
        ("click", (40, 12)),  # Right in the middle.
        #
        ("hover", (0, 0)),  # Top-left corner.
        ("hover", (40, 0)),  # Top edge.
        ("hover", (79, 0)),  # Top-right corner.
        ("hover", (79, 12)),  # Right edge.
        ("hover", (79, 23)),  # Bottom-right corner.
        ("hover", (40, 23)),  # Bottom edge.
        ("hover", (40, 23)),  # Bottom-left corner.
        ("hover", (0, 12)),  # Left edge.
        ("hover", (40, 12)),  # Right in the middle.
        #
        ("mouse_down", (0, 0)),  # Top-left corner.
        ("mouse_down", (40, 0)),  # Top edge.
        ("mouse_down", (79, 0)),  # Top-right corner.
        ("mouse_down", (79, 12)),  # Right edge.
        ("mouse_down", (79, 23)),  # Bottom-right corner.
        ("mouse_down", (40, 23)),  # Bottom edge.
        ("mouse_down", (40, 23)),  # Bottom-left corner.
        ("mouse_down", (0, 12)),  # Left edge.
        ("mouse_down", (40, 12)),  # Right in the middle.
        #
        ("mouse_up", (0, 0)),  # Top-left corner.
        ("mouse_up", (40, 0)),  # Top edge.
        ("mouse_up", (79, 0)),  # Top-right corner.
        ("mouse_up", (79, 12)),  # Right edge.
        ("mouse_up", (79, 23)),  # Bottom-right corner.
        ("mouse_up", (40, 23)),  # Bottom edge.
        ("mouse_up", (40, 23)),  # Bottom-left corner.
        ("mouse_up", (0, 12)),  # Left edge.
        ("mouse_up", (40, 12)),  # Right in the middle.
    ],
)
async def test_pilot_target_inside_screen_is_fine_with_correct_coordinate_system(
    method, offset
):
    """Make sure that the coordinate system for the click is the correct one.

    Especially relevant because I kept getting confused about the way it works.
    """
    app = ManyLabelsApp()
    async with app.run_test(size=(80, 24)) as pilot:
        app.query_one("#label99").scroll_visible(animate=False)
        await pilot.pause()

        pilot_method = getattr(pilot, method)
        await pilot_method(offset=offset)


@pytest.mark.parametrize(
    ["method", "target"],
    [
        ("click", "#label0"),
        ("click", "#label90"),
        ("click", Button),
        #
        ("hover", "#label0"),
        ("hover", "#label90"),
        ("hover", Button),
        #
        ("mouse_down", "#label0"),
        ("mouse_down", "#label90"),
        ("mouse_down", Button),
        #
        ("mouse_up", "#label0"),
        ("mouse_up", "#label90"),
        ("mouse_up", Button),
    ],
)
async def test_pilot_target_on_widget_that_is_not_visible_errors(method, target):
    """Make sure that clicking a widget that is not scrolled into view raises an error."""
    app = ManyLabelsApp()
    async with app.run_test(size=(80, 5)) as pilot:
        app.query_one("#label50").scroll_visible(animate=False)
        await pilot.pause()

        pilot_method = getattr(pilot, method)
        with pytest.raises(OutOfBounds):
            await pilot_method(target)


@pytest.mark.parametrize("method", ["click", "hover", "mouse_down", "mouse_up"])
async def test_pilot_target_widget_under_another_widget(method):
    """The targeting method should return False when the targeted widget is covered."""

    class ObscuredButton(App):
        CSS = """
        Label {
            width: 30;
            height: 5;
        }
        """

        def compose(self):
            yield Button()
            yield Label()

        def on_mount(self):
            self.query_one(Label).styles.offset = (0, -3)

    app = ObscuredButton()
    async with app.run_test() as pilot:
        await pilot.pause()
        pilot_method = getattr(pilot, method)
        assert (await pilot_method(Button)) is False


@pytest.mark.parametrize("method", ["click", "hover", "mouse_down", "mouse_up"])
async def test_pilot_target_visible_widget(method):
    """The targeting method should return True when the targeted widget is hit."""

    class ObscuredButton(App):
        def compose(self):
            yield Button()

    app = ObscuredButton()
    async with app.run_test() as pilot:
        await pilot.pause()
        pilot_method = getattr(pilot, method)
        assert (await pilot_method(Button)) is True


@pytest.mark.parametrize(
    ["method", "offset"],
    [
        ("click", (0, 0)),
        ("click", (2, 0)),
        ("click", (10, 23)),
        ("click", (70, 0)),
        #
        ("hover", (0, 0)),
        ("hover", (2, 0)),
        ("hover", (10, 23)),
        ("hover", (70, 0)),
        #
        ("mouse_down", (0, 0)),
        ("mouse_down", (2, 0)),
        ("mouse_down", (10, 23)),
        ("mouse_down", (70, 0)),
        #
        ("mouse_up", (0, 0)),
        ("mouse_up", (2, 0)),
        ("mouse_up", (10, 23)),
        ("mouse_up", (70, 0)),
    ],
)
async def test_pilot_target_screen_always_true(method, offset):
    app = ManyLabelsApp()
    async with app.run_test(size=(80, 24)) as pilot:
        pilot_method = getattr(pilot, method)
        assert (await pilot_method(offset=offset)) is True


async def test_pilot_resize_terminal():
    app = App()
    async with app.run_test(size=(80, 24)) as pilot:
        # Sanity checks.
        assert app.size == (80, 24)
        assert app.screen.size == (80, 24)
        await pilot.resize_terminal(27, 15)
        await pilot.pause()
        assert app.size == (27, 15)
        assert app.screen.size == (27, 15)


async def test_fail_early():
    # https://github.com/Textualize/textual/issues/3282
    class MyApp(App):
        CSS_PATH = "foo.tcss"

    app = MyApp()
    with pytest.raises(StylesheetError):
        async with app.run_test() as pilot:
            await pilot.press("enter")



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Header, Footer, Label
from textual.binding import Binding


class Dialog(Vertical):
    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        yield Label("This should not cause a scrollbar to appear")


class DialogIssueApp(App[None]):
    CSS = """
    Screen {
        layers: base dialog;
    }

    .hidden {
        display: none;
    }

    Dialog {
        align: center middle;
        border: round red;
        width: 50%;
        height: 50%;
        layer: dialog;
        offset: 50% 50%;
    }
    """

    BINDINGS = [
        Binding("d", "dialog", "Toggle the dialog"),
    ]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Vertical()
        yield Dialog(classes="hidden")
        yield Footer()

    def action_dialog(self) -> None:
        self.query_one(Dialog).toggle_class("hidden")


if __name__ == "__main__":
    DialogIssueApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Sparkline

data = [1, 2, 2, 1, 1, 4, 3, 1, 1, 8, 8, 2]  # (1)!


class SparklineBasicApp(App[None]):
    CSS_PATH = "sparkline_basic.tcss"

    def compose(self) -> ComposeResult:
        yield Sparkline(  # (2)!
            data,  # (3)!
            summary_function=max,  # (4)!
        )


app = SparklineBasicApp()
if __name__ == "__main__":
    app.run()



from fractions import Fraction

import pytest

from textual._resolve import resolve, resolve_fraction_unit
from textual.css.scalar import Scalar
from textual.geometry import Size
from textual.widget import Widget


def test_resolve_empty():
    assert resolve([], 10, 1, Size(20, 10), Size(80, 24)) == []


@pytest.mark.parametrize(
    "scalars,total,gutter,result",
    [
        (["10"], 100, 0, [(0, 10)]),
        (
            ["10", "20"],
            100,
            0,
            [(0, 10), (10, 20)],
        ),
        (
            ["10", "20"],
            100,
            1,
            [(0, 10), (11, 20)],
        ),
        (
            ["10", "1fr"],
            100,
            1,
            [(0, 10), (11, 89)],
        ),
        (
            ["1fr", "1fr"],
            100,
            0,
            [(0, 50), (50, 50)],
        ),
        (
            ["3", "1fr", "1fr", "1"],
            100,
            1,
            [(0, 3), (4, 46), (51, 47), (99, 1)],
        ),
    ],
)
def test_resolve(scalars, total, gutter, result):
    assert (
        resolve(
            [Scalar.parse(scalar) for scalar in scalars],
            total,
            gutter,
            Size(40, 20),
            Size(80, 24),
        )
        == result
    )


def test_resolve_fraction_unit():
    """Test resolving fraction units in combination with minimum widths."""
    widget1 = Widget()
    widget2 = Widget()
    widget3 = Widget()

    widget1.styles.width = "1fr"
    widget1.styles.min_width = 20

    widget2.styles.width = "2fr"
    widget2.styles.min_width = 10

    widget3.styles.width = "1fr"

    styles = (widget1.styles, widget2.styles, widget3.styles)

    # Try with width 80.
    # Fraction unit should one fourth of width
    assert resolve_fraction_unit(
        styles,
        Size(80, 24),
        Size(80, 24),
        Fraction(80),
        resolve_dimension="width",
    ) == Fraction(20)

    # Try with width 50
    # First widget is fixed at 20
    # Remaining three widgets have 30 to play with
    # Fraction is 10
    assert resolve_fraction_unit(
        styles,
        Size(80, 24),
        Size(80, 24),
        Fraction(50),
        resolve_dimension="width",
    ) == Fraction(10)

    # Try with width 35
    # First widget fixed at 20
    # Fraction is 5
    assert resolve_fraction_unit(
        styles,
        Size(80, 24),
        Size(80, 24),
        Fraction(35),
        resolve_dimension="width",
    ) == Fraction(5)

    # Try with width 32
    # First widget fixed at 20
    # Second widget is fixed at 10
    # Remaining widget has all the space
    # Fraction is 2
    assert resolve_fraction_unit(
        styles,
        Size(80, 24),
        Size(80, 24),
        Fraction(32),
        resolve_dimension="width",
    ) == Fraction(2)


def test_resolve_fraction_unit_stress_test():
    """Check for zero division errors."""
    # https://github.com/Textualize/textual/issues/2673
    widget = Widget()
    styles = widget.styles
    styles.width = "1fr"

    # We're mainly checking for the absence of zero division errors,
    # which is a reoccurring theme for this code.
    for remaining_space in range(1, 51, 10):
        for max_width in range(1, remaining_space):
            styles.max_width = max_width

            for width in range(1, remaining_space, 2):
                size = Size(width, 24)
                resolved_unit = resolve_fraction_unit(
                    [styles, styles, styles],
                    size,
                    size,
                    Fraction(remaining_space),
                    "width",
                )
                assert resolved_unit <= remaining_space


def test_resolve_issue_2502():
    """Test https://github.com/Textualize/textual/issues/2502"""

    widget = Widget()
    widget.styles.width = "1fr"
    widget.styles.min_width = 11

    assert isinstance(
        resolve_fraction_unit(
            (widget.styles,),
            Size(80, 24),
            Size(80, 24),
            Fraction(10),
            resolve_dimension="width",
        ),
        Fraction,
    )



import asyncio

import pytest

from textual.app import App
from textual.worker import (
    DeadlockError,
    NoActiveWorker,
    Worker,
    WorkerCancelled,
    WorkerError,
    WorkerFailed,
    WorkerState,
    get_current_worker,
)


async def test_initialize():
    """Test initial values."""

    def foo() -> str:
        return "foo"

    app = App()
    async with app.run_test():
        worker = Worker(app, foo, name="foo", group="foo-group", description="Foo test")
        repr(worker)

    assert worker.state == WorkerState.PENDING
    assert not worker.is_cancelled
    assert not worker.is_running
    assert not worker.is_finished
    assert worker.completed_steps == 0
    assert worker.total_steps is None
    assert worker.progress == 0.0
    assert worker.result is None


async def test_run_success() -> None:
    """Test successful runs."""

    def foo() -> str:
        """Regular function."""
        return "foo"

    async def bar() -> str:
        """Coroutine."""
        return "bar"

    async def baz() -> str:
        """Coroutine."""
        return "baz"

    class RunApp(App):
        pass

    app = RunApp()
    async with app.run_test():
        # Call regular function
        foo_worker: Worker[str] = Worker(
            app, foo, name="foo", group="foo-group", description="Foo test", thread=True
        )
        # Call coroutine function
        bar_worker: Worker[str] = Worker(
            app, bar, name="bar", group="bar-group", description="Bar test"
        )
        # Call coroutine
        baz_worker: Worker[str] = Worker(
            app, baz(), name="baz", group="baz-group", description="Baz test"
        )
        # Call coroutine function in a thread
        bar_thread_worker: Worker[str] = Worker(
            app, bar, name="bar", group="bar-group", description="Bar test", thread=True
        )
        # Call coroutine in a thread.
        baz_thread_worker: Worker[str] = Worker(
            app,
            baz(),
            name="baz",
            group="baz-group",
            description="Baz test",
            thread=True,
        )
        assert foo_worker.result is None
        assert bar_worker.result is None
        assert baz_worker.result is None
        assert bar_thread_worker.result is None
        assert baz_thread_worker.result is None
        foo_worker._start(app)
        bar_worker._start(app)
        baz_worker._start(app)
        bar_thread_worker._start(app)
        baz_thread_worker._start(app)
        assert await foo_worker.wait() == "foo"
        assert await bar_worker.wait() == "bar"
        assert await baz_worker.wait() == "baz"
        assert await bar_thread_worker.wait() == "bar"
        assert await baz_thread_worker.wait() == "baz"
        assert foo_worker.result == "foo"
        assert bar_worker.result == "bar"
        assert baz_worker.result == "baz"
        assert bar_thread_worker.result == "bar"
        assert baz_thread_worker.result == "baz"


async def test_run_error() -> None:
    async def run_error() -> str:
        await asyncio.sleep(0.1)
        1 / 0
        return "Never"

    class ErrorApp(App):
        pass

    app = ErrorApp()
    with pytest.raises(WorkerFailed):
        async with app.run_test():
            worker: Worker[str] = Worker(app, run_error)
            worker._start(app)
            await worker.wait()


async def test_run_cancel() -> None:
    """Test run may be cancelled."""

    async def run_error() -> str:
        await asyncio.sleep(0.1)
        return "Never"

    class ErrorApp(App):
        pass

    app = ErrorApp()
    async with app.run_test():
        worker: Worker[str] = Worker(app, run_error)
        worker._start(app)
        await asyncio.sleep(0)
        worker.cancel()
        assert worker.is_cancelled
        with pytest.raises(WorkerCancelled):
            await worker.wait()


async def test_run_cancel_immediately() -> None:
    """Edge case for cancelling immediately."""

    async def run_error() -> str:
        await asyncio.sleep(0.1)
        return "Never"

    class ErrorApp(App):
        pass

    app = ErrorApp()
    async with app.run_test():
        worker: Worker[str] = Worker(app, run_error)
        worker._start(app)
        worker.cancel()
        assert worker.is_cancelled
        with pytest.raises(WorkerCancelled):
            await worker.wait()


async def test_get_worker() -> None:
    """Check get current worker."""

    async def run_worker() -> Worker:
        worker = get_current_worker()
        return worker

    class WorkerApp(App):
        pass

    app = WorkerApp()
    async with app.run_test():
        worker: Worker[Worker] = Worker(app, run_worker)
        worker._start(app)

        assert await worker.wait() is worker


def test_no_active_worker() -> None:
    """No active worker raises a specific exception."""
    with pytest.raises(NoActiveWorker):
        get_current_worker()


async def test_progress_update():
    async def long_work():
        pass

    app = App()
    async with app.run_test():
        worker = Worker(app, long_work)
        worker._start(app)
        worker.update(total_steps=100)
        assert worker.progress == 0
        worker.advance(50)
        assert worker.progress == 50
        worker.update(completed_steps=23)
        assert worker.progress == 73


async def test_double_start():
    async def long_work():
        return 0

    app = App()
    async with app.run_test():
        worker = Worker(app, long_work)
        worker._start(app)
        worker._start(app)
        assert await worker.wait() == 0


async def test_self_referential_deadlock():
    async def self_referential_work():
        await get_current_worker().wait()

    app = App()
    with pytest.raises(WorkerFailed) as exc:
        async with app.run_test():
            worker = Worker(app, self_referential_work)
            worker._start(app)
            await worker.wait()
        assert exc.type is DeadlockError


async def test_wait_without_start():
    async def work():
        return

    app = App()
    async with app.run_test():
        worker = Worker(app, work)
        with pytest.raises(WorkerError):
            await worker.wait()



from textual.app import App
from textual.widgets import Label


class SelfRemovingLabel(Label):
    async def on_mount(self) -> None:
        await self.remove()


class RemoveOnTimerApp(App[None]):
    def on_mount(self):
        for _ in range(5):
            self.mount(SelfRemovingLabel("I will remove myself!"))


async def test_multiple_simultaneous_removals():
    """Regression test for https://github.com/Textualize/textual/issues/2854."""
    # The app should run and finish without raising any errors.
    async with RemoveOnTimerApp().run_test() as pilot:
        # Sanity check to ensure labels were removed.
        assert len(pilot.app.query(Label)) == 0



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import RadioButton, RadioSet


class RadioChoicesApp(App[None]):
    CSS_PATH = "radio_set.tcss"

    def compose(self) -> ComposeResult:
        with Horizontal():
            # A RadioSet built up from RadioButtons.
            with RadioSet(id="focus_me"):
                yield RadioButton("Battlestar Galactica")
                yield RadioButton("Dune 1984")
                yield RadioButton("Dune 2021")
                yield RadioButton("Serenity", value=True)
                yield RadioButton("Star Trek: The Motion Picture")
                yield RadioButton("Star Wars: A New Hope")
                yield RadioButton("The Last Starfighter")
                yield RadioButton(
                    "Total Recall :backhand_index_pointing_right: :red_circle:"
                )
                yield RadioButton("Wing Commander")
            # A RadioSet built up from a collection of strings.
            yield RadioSet(
                "Amanda",
                "Connor MacLeod",
                "Duncan MacLeod",
                "Heather MacLeod",
                "Joe Dawson",
                "Kurgan, [bold italic red]The[/]",
                "Methos",
                "Rachel Ellenstein",
                "Ramírez",
            )

    def on_mount(self) -> None:
        self.query_one("#focus_me").focus()


if __name__ == "__main__":
    RadioChoicesApp().run()



from typing import cast

from rich.align import Align, AlignMethod
from rich.console import (
    Console,
    ConsoleOptions,
    JustifyMethod,
    RenderableType,
    RenderResult,
)
from rich.measure import Measurement
from rich.segment import Segment, Segments
from rich.style import Style


class HorizontalPad:
    """Rich renderable to add padding on the left and right of a renderable.

    Note that unlike Rich's Padding class this align each line independently.

    """

    def __init__(
        self,
        renderable: RenderableType,
        left: int,
        right: int,
        pad_style: Style,
        justify: JustifyMethod,
    ) -> None:
        """
        Initialize HorizontalPad.

        Args:
            renderable: A Rich renderable.
            left: Left padding.
            right: Right padding.
            pad_style: Style of padding.
            justify: Justify method.
        """
        self.renderable = renderable
        self.left = left
        self.right = right
        self.pad_style = pad_style
        self.justify = justify

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        options = options.update(
            width=options.max_width - self.left - self.right, height=None
        )
        lines = console.render_lines(self.renderable, options, pad=False)
        left_pad = Segment(" " * self.left, self.pad_style)
        right_pad = Segment(" " * self.right, self.pad_style)

        align: AlignMethod = cast(
            AlignMethod,
            self.justify if self.justify in {"left", "right", "center"} else "left",
        )

        for line in lines:
            pad_line = line
            if self.left:
                pad_line = [left_pad, *line]
            if self.right:
                pad_line.append(right_pad)
            segments = Segments(pad_line)
            yield Align(segments, align=align)

    def __rich_measure__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> Measurement:
        measurement = Measurement.get(console, options, self.renderable)
        total_padding = self.left + self.right
        return Measurement(
            measurement.minimum + total_padding,
            measurement.maximum + total_padding,
        )



from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Header, Select

LINES = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.""".splitlines()

ALTERNATE_LINES = """Twinkle, twinkle, little star,
How I wonder what you are!
Up above the world so high,
Like a diamond in the sky.
Twinkle, twinkle, little star,
How I wonder what you are!""".splitlines()


class SelectApp(App):
    CSS_PATH = "select.tcss"

    BINDINGS = [("s", "swap", "Swap Select options")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Select(zip(LINES, LINES), allow_blank=False)

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed) -> None:
        self.title = str(event.value)

    def action_swap(self) -> None:
        self.query_one(Select).set_options(zip(ALTERNATE_LINES, ALTERNATE_LINES))


if __name__ == "__main__":
    app = SelectApp()
    app.run()



"""Core selection list unit tests, aimed at testing basic list creation.

Note that the vast majority of the API *isn't* tested in here as
`SelectionList` inherits from `OptionList` and so that would be duplicated
effort. Instead these tests aim to just test the things that have been
changed or wrapped in some way.
"""

from __future__ import annotations

import pytest
from rich.text import Text

from textual.app import App, ComposeResult
from textual.widgets import SelectionList
from textual.widgets.option_list import Option
from textual.widgets.selection_list import Selection, SelectionError


class SelectionListApp(App[None]):
    """Test selection list application."""

    def compose(self) -> ComposeResult:
        yield SelectionList[int](
            ("0", 0),
            ("1", 1, False),
            ("2", 2, True),
            Selection("3", 3, id="3"),
            Selection("4", 4, True, id="4"),
        )


async def test_all_parameters_become_selctions() -> None:
    """All input parameters to a list should become selections."""
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        assert selections.option_count == 5
        for n in range(5):
            assert isinstance(selections.get_option_at_index(n), Selection)


async def test_get_selection_by_index() -> None:
    """It should be possible to get a selection by index."""
    async with SelectionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(SelectionList)
        for n in range(5):
            assert option_list.get_option_at_index(n).prompt == Text(str(n))
        assert option_list.get_option_at_index(-1).prompt == Text("4")


async def test_get_selection_by_id() -> None:
    """It should be possible to get a selection by ID."""
    async with SelectionListApp().run_test() as pilot:
        option_list = pilot.app.query_one(SelectionList)
        assert option_list.get_option("3").prompt == Text("3")
        assert option_list.get_option("4").prompt == Text("4")


async def test_add_later() -> None:
    """It should be possible to add more items to a selection list."""
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        assert selections.option_count == 5
        selections.add_option(("5", 5))
        assert selections.option_count == 6
        selections.add_option(Selection("6", 6))
        assert selections.option_count == 7
        selections.add_options(
            [Selection("7", 7), Selection("8", 8, True), ("9", 9), ("10", 10, True)]
        )
        assert selections.option_count == 11
        selections.add_options([])
        assert selections.option_count == 11


async def test_add_later_selcted_state() -> None:
    """When adding selections later the selected collection should get updated."""
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        assert selections.selected == [2, 4]
        selections.add_option(("5", 5, True))
        assert selections.selected == [2, 4, 5]
        selections.add_option(Selection("6", 6, True))
        assert selections.selected == [2, 4, 5, 6]


async def test_add_non_selections() -> None:
    """Adding options that aren't selections should result in errors."""
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        with pytest.raises(SelectionError):
            selections.add_option(None)
        with pytest.raises(SelectionError):
            selections.add_option(Option("Nope"))
        with pytest.raises(SelectionError):
            selections.add_option("Nope")
        with pytest.raises(SelectionError):
            selections.add_option(("Nope",))
        with pytest.raises(SelectionError):
            selections.add_option(("Nope", 0, False, 23))


async def test_clear_options() -> None:
    """Clearing the options should also clear the selections."""
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        selections.clear_options()
        assert selections.selected == []


async def test_options_are_available_soon() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3903."""

    selection = Selection("", 0, id="some_id")
    selection_list = SelectionList[int](selection)
    assert selection_list.get_option("some_id") is selection


async def test_removing_option_updates_indexes() -> None:
    async with SelectionListApp().run_test() as pilot:
        selections = pilot.app.query_one(SelectionList)
        assert selections._values == {n: n for n in range(5)}

        selections.remove_option_at_index(0)
        assert selections._values == {n + 1: n for n in range(4)}



from textual import on
from textual.app import App
from textual.widgets import Select
from textual.widgets._select import SelectOverlay


class SelectApp(App[None]):
    def __init__(self):
        self.changed_messages = []
        super().__init__()

    def compose(self):
        yield Select[int]([(str(n), n) for n in range(3)])

    @on(Select.Changed)
    def add_message(self, event):
        self.changed_messages.append(event)


async def test_message_control():
    app = SelectApp()
    async with app.run_test() as pilot:
        await pilot.click(Select)
        await pilot.click(SelectOverlay, offset=(2, 3))
        await pilot.pause()
        message = app.changed_messages[0]
        assert message.control is app.query_one(Select)


async def test_selecting_posts_message():
    app = SelectApp()
    async with app.run_test() as pilot:
        await pilot.click(Select)
        # Click on the 1.
        await pilot.click(SelectOverlay, offset=(2, 3))
        await pilot.pause()
        assert len(app.changed_messages) == 1
        await pilot.click(Select)
        # Click on the 2.
        await pilot.click(SelectOverlay, offset=(2, 4))
        await pilot.pause()
        assert len(app.changed_messages) == 2


async def test_same_selection_does_not_post_message():
    app = SelectApp()
    async with app.run_test() as pilot:
        await pilot.click(Select)
        # Click on the 1.
        await pilot.click(SelectOverlay, offset=(2, 3))
        await pilot.pause()
        assert len(app.changed_messages) == 1
        await pilot.click(Select)
        # Click on the 1 again...
        await pilot.click(SelectOverlay, offset=(2, 3))
        await pilot.pause()
        assert len(app.changed_messages) == 1


async def test_setting_value_posts_message() -> None:
    """Setting the value of a Select should post a message."""

    async with (app := SelectApp()).run_test() as pilot:
        assert len(app.changed_messages) == 0
        app.query_one(Select).value = 2
        await pilot.pause()
        assert len(app.changed_messages) == 1



from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.events import Mount
from textual.widgets import Footer, Header, Pretty, SelectionList
from textual.widgets.selection_list import Selection


class SelectionListApp(App[None]):
    CSS_PATH = "selection_list_selected.tcss"

    def compose(self) -> ComposeResult:
        yield Header()
        with Horizontal():
            yield SelectionList[str](  # (1)!
                Selection("Falken's Maze", "secret_back_door", True),
                Selection("Black Jack", "black_jack"),
                Selection("Gin Rummy", "gin_rummy"),
                Selection("Hearts", "hearts"),
                Selection("Bridge", "bridge"),
                Selection("Checkers", "checkers"),
                Selection("Chess", "a_nice_game_of_chess", True),
                Selection("Poker", "poker"),
                Selection("Fighter Combat", "fighter_combat", True),
            )
            yield Pretty([])
        yield Footer()

    def on_mount(self) -> None:
        self.query_one(SelectionList).border_title = "Shall we play some games?"
        self.query_one(Pretty).border_title = "Selected games"

    @on(Mount)
    @on(SelectionList.SelectedChanged)
    def update_selected_view(self) -> None:
        self.query_one(Pretty).update(self.query_one(SelectionList).selected)


if __name__ == "__main__":
    SelectionListApp().run()



import pytest

from textual.widgets import Tree
from textual.widgets.tree import AddNodeError


def test_tree_node_add_before_and_after_raises_exception():
    tree = Tree[None]("root")
    with pytest.raises(AddNodeError):
        tree.root.add("error", before=99, after=0)


def test_tree_node_add_before_or_after_with_invalid_type_raises_exception():
    tree = Tree[None]("root")
    tree.root.add("node")
    with pytest.raises(TypeError):
        tree.root.add("before node", before="node")
    with pytest.raises(TypeError):
        tree.root.add("after node", after="node")


def test_tree_node_add_before_index():
    tree = Tree[None]("root")
    tree.root.add("node")
    tree.root.add("before node", before=0)
    tree.root.add("first", before=-99)
    tree.root.add("after first", before=-2)
    tree.root.add("last", before=99)
    tree.root.add("after node", before=4)
    tree.root.add("before last", before=-1)

    assert str(tree.root.children[0].label) == "first"
    assert str(tree.root.children[1].label) == "after first"
    assert str(tree.root.children[2].label) == "before node"
    assert str(tree.root.children[3].label) == "node"
    assert str(tree.root.children[4].label) == "after node"
    assert str(tree.root.children[5].label) == "before last"
    assert str(tree.root.children[6].label) == "last"


def test_tree_node_add_after_index():
    tree = Tree[None]("root")
    tree.root.add("node")
    tree.root.add("after node", after=0)
    tree.root.add("first", after=-99)
    tree.root.add("after first", after=-3)
    tree.root.add("before node", after=1)
    tree.root.add("before last", after=99)
    tree.root.add("last", after=-1)

    assert str(tree.root.children[0].label) == "first"
    assert str(tree.root.children[1].label) == "after first"
    assert str(tree.root.children[2].label) == "before node"
    assert str(tree.root.children[3].label) == "node"
    assert str(tree.root.children[4].label) == "after node"
    assert str(tree.root.children[5].label) == "before last"
    assert str(tree.root.children[6].label) == "last"


def test_tree_node_add_relative_to_unknown_node_raises_exception():
    tree = Tree[None]("root")
    removed_node = tree.root.add("removed node")
    removed_node.remove()
    with pytest.raises(AddNodeError):
        tree.root.add("node", before=removed_node)
    with pytest.raises(AddNodeError):
        tree.root.add("node", after=removed_node)


def test_tree_node_add_before_node():
    tree = Tree[None]("root")
    node = tree.root.add("node")
    before_node = tree.root.add("before node", before=node)
    tree.root.add("first", before=before_node)
    tree.root.add("after first", before=before_node)
    last = tree.root.add("last", before=4)
    before_last = tree.root.add("before last", before=last)
    tree.root.add("after node", before=before_last)

    assert str(tree.root.children[0].label) == "first"
    assert str(tree.root.children[1].label) == "after first"
    assert str(tree.root.children[2].label) == "before node"
    assert str(tree.root.children[3].label) == "node"
    assert str(tree.root.children[4].label) == "after node"
    assert str(tree.root.children[5].label) == "before last"
    assert str(tree.root.children[6].label) == "last"


def test_tree_node_add_after_node():
    tree = Tree[None]("root")
    node = tree.root.add("node")
    after_node = tree.root.add("after node", after=node)
    first = tree.root.add("first", after=-3)
    after_first = tree.root.add("after first", after=first)
    tree.root.add("before node", after=after_first)
    before_last = tree.root.add("before last", after=after_node)
    tree.root.add("last", after=before_last)

    assert str(tree.root.children[0].label) == "first"
    assert str(tree.root.children[1].label) == "after first"
    assert str(tree.root.children[2].label) == "before node"
    assert str(tree.root.children[3].label) == "node"
    assert str(tree.root.children[4].label) == "after node"
    assert str(tree.root.children[5].label) == "before last"
    assert str(tree.root.children[6].label) == "last"


def test_tree_node_add_leaf_before_or_after():
    tree = Tree[None]("root")
    leaf = tree.root.add_leaf("leaf")
    tree.root.add_leaf("before leaf", before=leaf)
    tree.root.add_leaf("after leaf", after=leaf)
    tree.root.add_leaf("first", before=0)
    tree.root.add_leaf("last", after=-1)

    assert str(tree.root.children[0].label) == "first"
    assert str(tree.root.children[1].label) == "before leaf"
    assert str(tree.root.children[2].label) == "leaf"
    assert str(tree.root.children[3].label) == "after leaf"
    assert str(tree.root.children[4].label) == "last"



from textual.app import App, ComposeResult
from textual.events import Resize
from textual.widgets import RichLog

from rich.text import Text


class RichLogWidth(App[None]):
    def compose(self) -> ComposeResult:
        rich_log = RichLog(min_width=20)
        rich_log.display = False
        rich_log.write(
            Text("written in compose", style="black on white", justify="right"),
            expand=True,
        )
        yield rich_log

    def key_p(self, event: Resize) -> None:
        rich_log: RichLog = self.query_one(RichLog)
        rich_log.display = True
        rich_log.write(Text("hello1", style="on red", justify="right"), expand=True)
        rich_log.visible = False
        rich_log.write(Text("world2", style="on green", justify="right"), expand=True)
        rich_log.visible = True
        rich_log.write(Text("hello3", style="on blue", justify="right"), expand=True)
        rich_log.display = False
        rich_log.write(Text("world4", style="on yellow", justify="right"), expand=True)
        rich_log.display = True


app = RichLogWidth()
if __name__ == "__main__":
    app.run()



from textual.app import App
from textual.command import CommandList, CommandPalette, Hit, Hits, Provider


class SimpleSource(Provider):
    async def search(self, query: str) -> Hits:
        def goes_nowhere_does_nothing() -> None:
            pass

        for _ in range(100):
            yield Hit(1, query, goes_nowhere_does_nothing, query)


class CommandPaletteApp(App[None]):
    COMMANDS = {SimpleSource}

    def on_mount(self) -> None:
        self.action_command_palette()


async def test_initial_list_no_highlight() -> None:
    """When the list initially appears, nothing will be highlighted."""
    async with CommandPaletteApp().run_test() as pilot:
        assert CommandPalette.is_open(pilot.app)
        assert pilot.app.screen.query_one(CommandList).visible is False
        await pilot.press("a")
        assert pilot.app.screen.query_one(CommandList).visible is True
        assert pilot.app.screen.query_one(CommandList).highlighted is None


async def test_down_arrow_selects_an_item() -> None:
    """Typing in a search value then pressing down should select a command."""
    async with CommandPaletteApp().run_test() as pilot:
        assert CommandPalette.is_open(pilot.app)
        assert pilot.app.screen.query_one(CommandList).visible is False
        await pilot.press("a")
        assert pilot.app.screen.query_one(CommandList).visible is True
        assert pilot.app.screen.query_one(CommandList).highlighted is None
        await pilot.press("down")
        assert pilot.app.screen.query_one(CommandList).highlighted is not None


async def test_enter_selects_an_item() -> None:
    """Typing in a search value then pressing enter should select a command."""
    async with CommandPaletteApp().run_test() as pilot:
        assert CommandPalette.is_open(pilot.app)
        assert pilot.app.screen.query_one(CommandList).visible is False
        await pilot.press("a")
        assert pilot.app.screen.query_one(CommandList).visible is True
        assert pilot.app.screen.query_one(CommandList).highlighted is None
        await pilot.press("enter")
        assert pilot.app.screen.query_one(CommandList).highlighted is not None


async def test_selection_of_command_closes_command_palette() -> None:
    """Selecting a command from the list should close the list."""
    async with CommandPaletteApp().run_test() as pilot:
        assert CommandPalette.is_open(pilot.app)
        assert pilot.app.screen.query_one(CommandList).visible is False
        await pilot.press("a")
        assert pilot.app.screen.query_one(CommandList).visible is True
        assert pilot.app.screen.query_one(CommandList).highlighted is None
        await pilot.press("enter")
        assert pilot.app.screen.query_one(CommandList).highlighted is not None
        await pilot.press("enter")
        assert not CommandPalette.is_open(pilot.app)



from __future__ import annotations


from textual.app import App, ComposeResult
from textual.widgets import OptionList, Header, Footer
from textual.widgets.option_list import Option


class OptionListApp(App[None]):

    def compose(self) -> ComposeResult:
        yield Header()
        yield OptionList(
            Option("1. Single line", id="one"),
            Option("2. Two\nlines", id="two"),
            Option("3. Three\nlines\nof text", id="three"),
        )

        yield Footer()

    def key_1(self):
        self.query_one(OptionList).replace_option_prompt_at_index(0, "1. Another single line")

    def key_2(self):
        self.query_one(OptionList).replace_option_prompt_at_index(0, "1. Two\nlines")

    def key_3(self):
        self.query_one(OptionList).replace_option_prompt_at_index(1, "1. Three\nlines\nof text")


if __name__ == "__main__":
    OptionListApp().run()



from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.widgets import Input, Label


class InputWithLabel(Widget):
    """An input with a label."""

    DEFAULT_CSS = """
    InputWithLabel {
        layout: horizontal;
        height: auto;
    }
    InputWithLabel Label {
        padding: 1;
        width: 12;
        text-align: right;
    }
    InputWithLabel Input {
        width: 1fr;
    }
    """

    def __init__(self, input_label: str) -> None:
        self.input_label = input_label
        super().__init__()

    def compose(self) -> ComposeResult:  # (1)!
        yield Label(self.input_label)
        yield Input()


class CompoundApp(App):
    CSS = """
    Screen {
        align: center middle;
    }
    InputWithLabel {
        width: 80%;
        margin: 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield InputWithLabel("First Name")
        yield InputWithLabel("Last Name")
        yield InputWithLabel("Email")


if __name__ == "__main__":
    app = CompoundApp()
    app.run()



import contextlib

from rich.terminal_theme import DIMMED_MONOKAI, MONOKAI, NIGHT_OWLISH

from textual.app import App, ComposeResult
from textual.widgets import Button, Input, Static


def test_batch_update():
    """Test `batch_update` context manager"""
    app = App()
    assert app._batch_count == 0  # Start at zero

    with app.batch_update():
        assert app._batch_count == 1  # Increments in context manager

        with app.batch_update():
            assert app._batch_count == 2  # Nested updates

        assert app._batch_count == 1  # Exiting decrements

    assert app._batch_count == 0  # Back to zero


class MyApp(App):
    def compose(self) -> ComposeResult:
        yield Input()
        yield Button("Click me!")


async def test_hover_update_styles():
    app = MyApp(ansi_color=False)
    async with app.run_test() as pilot:
        button = app.query_one(Button)
        assert button.pseudo_classes == {
            "blur",
            "can-focus",
            "dark",
            "enabled",
        }

        # Take note of the initial background colour
        initial_background = button.styles.background
        await pilot.hover(Button)

        # We've hovered, so ensure the pseudoclass is present and background changed
        assert button.pseudo_classes == {
            "blur",
            "can-focus",
            "dark",
            "enabled",
            "hover",
        }
        assert button.styles.background != initial_background


def test_setting_title():
    app = MyApp()
    app.title = None
    assert app.title == "None"

    app.title = ""
    assert app.title == ""

    app.title = 0.125
    assert app.title == "0.125"

    app.title = [True, False, 2]
    assert app.title == "[True, False, 2]"


def test_setting_sub_title():
    app = MyApp()
    app.sub_title = None
    assert app.sub_title == "None"

    app.sub_title = ""
    assert app.sub_title == ""

    app.sub_title = 0.125
    assert app.sub_title == "0.125"

    app.sub_title = [True, False, 2]
    assert app.sub_title == "[True, False, 2]"


async def test_default_return_code_is_zero():
    app = App()
    async with app.run_test():
        app.exit()
    assert app.return_code == 0


async def test_return_code_is_one_after_crash():
    class MyApp(App):
        def key_p(self):
            1 / 0

    app = MyApp()
    with contextlib.suppress(ZeroDivisionError):
        async with app.run_test() as pilot:
            await pilot.press("p")
    assert app.return_code == 1


async def test_set_return_code():
    app = App()
    async with app.run_test():
        app.exit(return_code=42)
    assert app.return_code == 42


def test_no_return_code_before_running():
    app = App()
    assert app.return_code is None


async def test_no_return_code_while_running():
    app = App()
    async with app.run_test():
        assert app.return_code is None


async def test_ansi_theme():
    app = App()
    async with app.run_test():
        app.ansi_theme_dark = NIGHT_OWLISH
        assert app.ansi_theme == NIGHT_OWLISH

        app.dark = False
        assert app.ansi_theme != NIGHT_OWLISH

        app.ansi_theme_light = MONOKAI
        assert app.ansi_theme == MONOKAI

        # Ensure if we change the dark theme while on light mode,
        # then change back to dark mode, the dark theme is updated.
        app.ansi_theme_dark = DIMMED_MONOKAI
        assert app.ansi_theme == MONOKAI

        app.dark = True
        assert app.ansi_theme == DIMMED_MONOKAI


async def test_early_exit():
    """Test exiting early doesn't cause issues."""
    from textual.app import App

    class AppExit(App):
        def compose(self):
            yield Static("Hello")

        def on_mount(self) -> None:
            # Exit after creating app
            self.exit()

    app = AppExit()
    async with app.run_test():
        pass


def test_early_exit_inline():
    """Test exiting early in inline mode doesn't break."""
    from textual.app import App

    class AppExit(App):
        def compose(self):
            yield Static("Hello")

        def on_mount(self) -> None:
            # Exit after creating app
            self.exit()

    app = AppExit()
    app.run(inline=True, inline_no_clear=True)



from textual.app import App, ComposeResult
from textual.widgets import Label, ListItem, ListView


class ListViewDisabledItemsApp(App[None]):
    def compose(self) -> ComposeResult:
        self.highlighted = []
        yield ListView(
            ListItem(Label("0"), disabled=True),
            ListItem(Label("1")),
            ListItem(Label("2"), disabled=True),
            ListItem(Label("3"), disabled=True),
            ListItem(Label("4")),
            ListItem(Label("5")),
            ListItem(Label("6"), disabled=True),
            ListItem(Label("7")),
            ListItem(Label("8"), disabled=True),
        )

    def _on_list_view_highlighted(self, message: ListView.Highlighted) -> None:
        if message.item is None:
            self.highlighted.append(None)
        else:
            self.highlighted.append(str(message.item.children[0].renderable))


async def test_keyboard_navigation_with_disabled_items() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/3881."""

    app = ListViewDisabledItemsApp()
    async with app.run_test() as pilot:
        for _ in range(5):
            await pilot.press("down")
        for _ in range(5):
            await pilot.press("up")

    assert app.highlighted == [
        "1",
        "4",
        "5",
        "7",
        "5",
        "4",
        "1",
    ]



import pytest

from textual.app import App
from textual.widgets import Select
from textual.widgets.select import InvalidSelectValueError

SELECT_OPTIONS = [(str(n), n) for n in range(3)]


async def test_value_is_blank_by_default():
    class SelectApp(App[None]):
        def compose(self):
            yield Select(SELECT_OPTIONS)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == Select.BLANK
        assert select.is_blank()


async def test_setting_and_checking_blank():
    class SelectApp(App[None]):
        def compose(self):
            yield Select(SELECT_OPTIONS)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == Select.BLANK
        assert select.is_blank()

        select.value = 0
        assert select.value != Select.BLANK
        assert not select.is_blank()

        select.value = Select.BLANK
        assert select.value == Select.BLANK
        assert select.is_blank()


async def test_clear_with_allow_blanks():
    class SelectApp(App[None]):
        def compose(self):
            yield Select(SELECT_OPTIONS, value=1)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert select.value == 1  # Sanity check.
        select.clear()
        assert select.is_blank()


async def test_clear_fails_if_allow_blank_is_false():
    class SelectApp(App[None]):
        def compose(self):
            yield Select(SELECT_OPTIONS, allow_blank=False)

    app = SelectApp()
    async with app.run_test():
        select = app.query_one(Select)
        assert not select.is_blank()
        with pytest.raises(InvalidSelectValueError):
            select.clear()



from textual.app import App, ComposeResult
from textual.widgets import TextArea

TEXT = """\
Foo
Bar
Baz
"""


class LineNumbersReactive(App[None]):
    START_LINE_NUMBER = 9999

    def compose(self) -> ComposeResult:
        yield TextArea(
            TEXT,
            soft_wrap=True,
            show_line_numbers=True,
            line_number_start=self.START_LINE_NUMBER,
        )


app = LineNumbersReactive()
if __name__ == "__main__":
    app.run()



from __future__ import annotations

from functools import partial
from pathlib import Path

from textual.app import App, ComposeResult
from textual.command import Hit, Hits, Provider
from textual.containers import VerticalScroll
from textual.widgets import Static


class PythonFileCommands(Provider):
    """A command provider to open a Python file in the current working directory."""

    def read_files(self) -> list[Path]:
        """Get a list of Python files in the current working directory."""
        return list(Path("./").glob("*.py"))

    async def startup(self) -> None:  # (1)!
        """Called once when the command palette is opened, prior to searching."""
        worker = self.app.run_worker(self.read_files, thread=True)
        self.python_paths = await worker.wait()

    async def search(self, query: str) -> Hits:  # (2)!
        """Search for Python files."""
        matcher = self.matcher(query)  # (3)!

        app = self.app
        assert isinstance(app, ViewerApp)

        for path in self.python_paths:
            command = f"open {str(path)}"
            score = matcher.match(command)  # (4)!
            if score > 0:
                yield Hit(
                    score,
                    matcher.highlight(command),  # (5)!
                    partial(app.open_file, path),
                    help="Open this file in the viewer",
                )


class ViewerApp(App):
    """Demonstrate a command source."""

    COMMANDS = App.COMMANDS | {PythonFileCommands}  # (6)!

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            yield Static(id="code", expand=True)

    def open_file(self, path: Path) -> None:
        """Open and display a file with syntax highlighting."""
        from rich.syntax import Syntax

        syntax = Syntax.from_path(
            str(path),
            line_numbers=True,
            word_wrap=False,
            indent_guides=True,
            theme="github-dark",
        )
        self.query_one("#code", Static).update(syntax)


if __name__ == "__main__":
    app = ViewerApp()
    app.run()



"""Provides a RadioSet widget, which groups radio buttons."""

from __future__ import annotations

from typing import ClassVar, Optional

import rich.repr
from rich.console import RenderableType

from textual import _widget_navigation
from textual.binding import Binding, BindingType
from textual.containers import Container
from textual.events import Click, Mount
from textual.message import Message
from textual.reactive import var
from textual.widgets._radio_button import RadioButton


class RadioSet(Container, can_focus=True, can_focus_children=False):
    """Widget for grouping a collection of radio buttons into a set.

    When a collection of [`RadioButton`][textual.widgets.RadioButton]s are
    grouped with this widget, they will be treated as a mutually-exclusive
    grouping. If one button is turned on, the previously-on button will be
    turned off.
    """

    DEFAULT_CSS = """
    RadioSet {
        border: tall transparent;
        background: $boost;
        padding: 0 1 0 0;
        height: auto;
        width: auto;
    }

    RadioSet:focus {
        border: tall $accent;
    }

    /* The following rules/styles mimic similar ToggleButton:focus rules in
     * ToggleButton. If those styles ever get updated, these should be too.
     */

    RadioSet > * {
        background: transparent;
        border: none;
        padding: 0 1;
    }

    RadioSet:focus > RadioButton.-selected > .toggle--label {
        text-style: underline;
    }

    RadioSet:focus ToggleButton.-selected > .toggle--button {
        background: $foreground 25%;
    }

    RadioSet:focus > RadioButton.-on.-selected > .toggle--button {
        background: $foreground 25%;
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("down,right", "next_button", "", show=False),
        Binding("enter,space", "toggle_button", "Toggle", show=False),
        Binding("up,left", "previous_button", "", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | enter, space | Toggle the currently-selected button. |
    | left, up | Select the previous radio button in the set. |
    | right, down | Select the next radio button in the set. |
    """

    _selected: var[int | None] = var[Optional[int]](None)
    """The index of the currently-selected radio button."""

    @rich.repr.auto
    class Changed(Message):
        """Posted when the pressed button in the set changes.

        This message can be handled using an `on_radio_set_changed` method.
        """

        ALLOW_SELECTOR_MATCH = {"pressed"}
        """Additional message attributes that can be used with the [`on` decorator][textual.on]."""

        def __init__(self, radio_set: RadioSet, pressed: RadioButton) -> None:
            """Initialise the message.

            Args:
                pressed: The radio button that was pressed.
            """
            super().__init__()
            self.radio_set = radio_set
            """A reference to the [`RadioSet`][textual.widgets.RadioSet] that was changed."""
            self.pressed = pressed
            """The [`RadioButton`][textual.widgets.RadioButton] that was pressed to make the change."""
            self.index = radio_set.pressed_index
            """The index of the [`RadioButton`][textual.widgets.RadioButton] that was pressed to make the change."""

        @property
        def control(self) -> RadioSet:
            """A reference to the [`RadioSet`][textual.widgets.RadioSet] that was changed.

            This is an alias for [`Changed.radio_set`][textual.widgets.RadioSet.Changed.radio_set]
            and is used by the [`on`][textual.on] decorator.
            """
            return self.radio_set

        def __rich_repr__(self) -> rich.repr.Result:
            yield "radio_set", self.radio_set
            yield "pressed", self.pressed
            yield "index", self.index

    def __init__(
        self,
        *buttons: str | RadioButton,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
    ) -> None:
        """Initialise the radio set.

        Args:
            buttons: The labels or [`RadioButton`][textual.widgets.RadioButton]s to group together.
            name: The name of the radio set.
            id: The ID of the radio set in the DOM.
            classes: The CSS classes of the radio set.
            disabled: Whether the radio set is disabled or not.
            tooltip: Optional tooltip.

        Note:
            When a `str` label is provided, a
            [RadioButton][textual.widgets.RadioButton] will be created from
            it.
        """
        self._pressed_button: RadioButton | None = None
        """Holds the radio buttons we're responsible for."""
        super().__init__(
            *[
                (button if isinstance(button, RadioButton) else RadioButton(button))
                for button in buttons
            ],
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        if tooltip is not None:
            self.tooltip = tooltip

    def _on_mount(self, _: Mount) -> None:
        """Perform some processing once mounted in the DOM."""

        # If there are radio buttons, select the first available one.
        self.action_next_button()

        # Get all the buttons within us; we'll be doing a couple of things
        # with that list.
        buttons = list(self.query(RadioButton))

        # RadioButtons can have focus, by default. But we're going to take
        # that over and handle movement between them. So here we tell them
        # all they can't focus.
        for button in buttons:
            button.can_focus = False

        # It's possible for the user to pass in a collection of radio
        # buttons, with more than one set to on; they shouldn't, but we
        # can't stop them. So here we check for that and, for want of a
        # better approach, we keep the first one on and turn all the others
        # off.
        switched_on = [button for button in buttons if button.value]
        with self.prevent(RadioButton.Changed):
            for button in switched_on[1:]:
                button.value = False

        # Keep track of which button is initially pressed.
        if switched_on:
            self._pressed_button = switched_on[0]

    def watch__selected(self) -> None:
        self.query(RadioButton).remove_class("-selected")
        if self._selected is not None:
            self._nodes[self._selected].add_class("-selected")

    def _on_radio_button_changed(self, event: RadioButton.Changed) -> None:
        """Respond to the value of a button in the set being changed.

        Args:
            event: The event.
        """
        # We're going to consume the underlying radio button events, making
        # it appear as if they don't emit their own, as far as the caller is
        # concerned. As such, stop the event bubbling and also prohibit the
        # same event being sent out if/when we make a value change in here.
        event.stop()
        with self.prevent(RadioButton.Changed):
            # If the message pertains to a button being clicked to on...
            if event.radio_button.value:
                # If there's a button pressed right now and it's not really a
                # case of the user mashing on the same button...
                if (
                    self._pressed_button is not None
                    and self._pressed_button != event.radio_button
                ):
                    self._pressed_button.value = False
                # Make the pressed button this new button.
                self._pressed_button = event.radio_button
                # Emit a message to say our state has changed.
                self.post_message(self.Changed(self, event.radio_button))
            else:
                # We're being clicked off, we don't want that.
                event.radio_button.value = True

    def _on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        """Handle a change to which button in the set is pressed.

        This handler ensures that, when a button is pressed, it's also the
        selected button.
        """
        self._selected = event.index

    async def _on_click(self, _: Click) -> None:
        """Handle a click on or within the radio set.

        This handler ensures that focus moves to the clicked radio set, even
        if there's a click on one of the radio buttons it contains.
        """
        self.focus()

    @property
    def pressed_button(self) -> RadioButton | None:
        """The currently-pressed [`RadioButton`][textual.widgets.RadioButton], or `None` if none are pressed."""
        return self._pressed_button

    @property
    def pressed_index(self) -> int:
        """The index of the currently-pressed [`RadioButton`][textual.widgets.RadioButton], or -1 if none are pressed."""
        return (
            self._nodes.index(self._pressed_button)
            if self._pressed_button is not None
            else -1
        )

    def action_previous_button(self) -> None:
        """Navigate to the previous button in the set.

        Note that this will wrap around to the end if at the start.
        """
        self._selected = _widget_navigation.find_next_enabled(
            self.children,
            anchor=self._selected,
            direction=-1,
        )

    def action_next_button(self) -> None:
        """Navigate to the next button in the set.

        Note that this will wrap around to the start if at the end.
        """
        self._selected = _widget_navigation.find_next_enabled(
            self.children,
            anchor=self._selected,
            direction=1,
        )

    def action_toggle_button(self) -> None:
        """Toggle the state of the currently-selected button."""
        if self._selected is not None:
            button = self._nodes[self._selected]
            assert isinstance(button, RadioButton)
            button.toggle()



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Label


class SuspendKeysApp(App[None]):

    BINDINGS = [Binding("ctrl+z", "suspend_process")]

    def compose(self) -> ComposeResult:
        yield Label("Press Ctrl+Z to suspend!")


if __name__ == "__main__":
    SuspendKeysApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Welcome

class WelcomeApp(App[None]):

    def compose(self) -> ComposeResult:
        yield Welcome()

if __name__ == "__main__":
    WelcomeApp().run()



from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Label, Static


# Checks  https://github.com/Textualize/textual/issues/2181
class MyCustomWidget(Static):
    def compose(self) -> ComposeResult:
        yield Label(("|\n" * 100)[:-1])
        yield Label("SHOULD BE VISIBLE", id="target")


class MyApp(App):
    def compose(self) -> ComposeResult:
        with VerticalScroll():
            yield MyCustomWidget()

    def key_t(self) -> None:
        self.query_one("#target").scroll_visible()


if __name__ == "__main__":
    MyApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Button


class ButtonIssue(App[None]):
    AUTO_FOCUS = None
    CSS = """
    Button {
        outline: white;
    }
    """

    def compose(self) -> ComposeResult:
        yield Button("Test")


if __name__ == "__main__":
    ButtonIssue().run()



from textual.app import App
from textual.widget import Widget


class WidthApp(App):
    CSS_PATH = "width.tcss"

    def compose(self):
        yield Widget()


if __name__ == "__main__":
    app = WidthApp()
    app.run()



from __future__ import annotations

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import OptionList
from textual.widgets.option_list import Option, Separator


class OptionListApp(App[None]):
    messages: list[str] = []

    def compose(self) -> ComposeResult:
        yield OptionList(
            Option("0"),
            Separator(),
            Option("1"),
        )

    @on(OptionList.OptionMessage)
    def log_option_message(
        self,
        event: OptionList.OptionMessage,
    ) -> None:
        self.messages.append(event.__class__.__name__)


async def test_option_list_clicking_separator() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/4710"""
    app = OptionListApp()
    async with app.run_test() as pilot:
        option_list = pilot.app.query_one(OptionList)
        expected_messages = ["OptionHighlighted"]  # Initial highlight
        assert option_list.highlighted == 0
        assert app.messages == expected_messages

        # Select the second option with the mouse
        await pilot.click(OptionList, offset=(3, 3))
        expected_messages.extend(["OptionHighlighted", "OptionSelected"])
        assert option_list.highlighted == 1
        assert app.messages == expected_messages

        # Click the separator - there should be no change
        await pilot.click(OptionList, offset=(3, 2))
        assert option_list.highlighted == 1
        assert app.messages == expected_messages



from operator import attrgetter

import pytest
from rich.text import Text

from textual import events
from textual._node_list import DuplicateIds
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.css.errors import StyleValueError
from textual.css.query import NoMatches
from textual.geometry import Offset, Size
from textual.message import Message
from textual.widget import BadWidgetName, MountError, PseudoClasses, Widget
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Input,
    Label,
    LoadingIndicator,
    Log,
    OptionList,
    RichLog,
    Switch,
    TextArea,
)


async def test_widget_construct():
    """Regression test for https://github.com/Textualize/textual/issues/5042"""

    # Check that constructing the widget outside of the app, doesn't invoke code that
    # expects an active app.
    class MyApp(App):
        def __init__(self) -> None:
            super().__init__()
            self.button = Button()
            self.data_table = DataTable()
            self.footer = Footer()
            self.header = Header()
            self.input = Input()
            self.label = Label()
            self.loading_indicator = LoadingIndicator()
            self.log_ = Log()
            self.option_list = OptionList()
            self.rich_log = RichLog()
            self.switch = Switch()
            self.text_area = TextArea(language="python")

    app = MyApp()
    async with app.run_test():
        pass


@pytest.mark.parametrize(
    "set_val, get_val, style_str",
    [
        [True, True, "visible"],
        [False, False, "hidden"],
        ["hidden", False, "hidden"],
        ["visible", True, "visible"],
    ],
)
def test_widget_set_visible_true(set_val, get_val, style_str):
    widget = Widget()
    widget.visible = set_val

    assert widget.visible is get_val
    assert widget.styles.visibility == style_str


def test_widget_set_visible_invalid_string():
    widget = Widget()

    with pytest.raises(StyleValueError):
        widget.visible = "nope! no widget for me!"

    assert widget.visible


def test_widget_content_width():
    class TextWidget(Widget):
        def __init__(self, text: str, id: str) -> None:
            self.text = text
            super().__init__(id=id)
            self.expand = False
            self.shrink = True

        def render(self) -> str:
            return self.text

    widget1 = TextWidget("foo", id="widget1")
    widget2 = TextWidget("foo\nbar", id="widget2")
    widget3 = TextWidget("foo\nbar\nbaz", id="widget3")

    app = App()
    with app._context():
        width = widget1.get_content_width(Size(20, 20), Size(80, 24))
        height = widget1.get_content_height(Size(20, 20), Size(80, 24), width)
        assert width == 3
        assert height == 1

        width = widget2.get_content_width(Size(20, 20), Size(80, 24))
        height = widget2.get_content_height(Size(20, 20), Size(80, 24), width)
        assert width == 3
        assert height == 2

        width = widget3.get_content_width(Size(20, 20), Size(80, 24))
        height = widget3.get_content_height(Size(20, 20), Size(80, 24), width)
        assert width == 3
        assert height == 3


class GetByIdApp(App):
    def compose(self) -> ComposeResult:
        grandchild1 = Widget(id="grandchild1")
        child1 = Widget(grandchild1, id="child1")
        child2 = Widget(id="child2")

        yield Widget(
            child1,
            child2,
            id="parent",
        )

    @property
    def parent(self) -> Widget:
        return self.query_one("#parent")


@pytest.fixture
async def hierarchy_app():
    app = GetByIdApp()
    yield app


async def test_get_child_by_id_gets_first_child(hierarchy_app):
    async with hierarchy_app.run_test():
        parent = hierarchy_app.parent
        child = parent.get_child_by_id(id="child1")
        assert child.id == "child1"
        assert child.get_child_by_id(id="grandchild1").id == "grandchild1"
        assert parent.get_child_by_id(id="child2").id == "child2"


async def test_get_child_by_id_no_matching_child(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        with pytest.raises(NoMatches):
            parent.get_child_by_id(id="doesnt-exist")


async def test_get_child_by_id_only_immediate_descendents(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        with pytest.raises(NoMatches):
            parent.get_child_by_id(id="grandchild1")


async def test_get_child_by_type():
    class GetChildApp(App):
        def compose(self) -> ComposeResult:
            yield Widget(id="widget1")
            yield Container(
                Label(id="label1"),
                Widget(id="widget2"),
                id="container1",
            )

    app = GetChildApp()
    async with app.run_test():
        assert app.get_child_by_type(Widget).id == "widget1"
        assert app.get_child_by_type(Container).id == "container1"
        with pytest.raises(NoMatches):
            app.get_child_by_type(Label)


async def test_get_widget_by_id_no_matching_child(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        with pytest.raises(NoMatches):
            parent.get_widget_by_id(id="i-dont-exist")


async def test_get_widget_by_id_non_immediate_descendants(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        result = parent.get_widget_by_id("grandchild1")
        assert result.id == "grandchild1"


async def test_get_widget_by_id_immediate_descendants(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        result = parent.get_widget_by_id("child1")
        assert result.id == "child1"


async def test_get_widget_by_id_doesnt_return_self(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        with pytest.raises(NoMatches):
            parent.get_widget_by_id("parent")


async def test_get_widgets_app_delegated(hierarchy_app):
    # Check that get_child_by_id finds the parent, which is a child of the default Screen
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        queried_parent = hierarchy_app.get_child_by_id("parent")
        assert queried_parent is parent

        # Check that the grandchild (descendant of the default screen) is found
        grandchild = hierarchy_app.get_widget_by_id("grandchild1")
        assert grandchild.id == "grandchild1"


async def test_widget_mount_ids_must_be_unique_mounting_all_in_one_go(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        widget1 = Widget(id="hello")
        widget2 = Widget(id="hello")

        with pytest.raises(MountError):
            parent.mount(widget1, widget2)


async def test_widget_mount_ids_must_be_unique_mounting_multiple_calls(hierarchy_app):
    async with hierarchy_app.run_test() as pilot:
        parent = pilot.app.parent
        widget1 = Widget(id="hello")
        widget2 = Widget(id="hello")

        parent.mount(widget1)
        with pytest.raises(DuplicateIds):
            parent.mount(widget2)


def test_get_pseudo_class_state():
    widget = Widget()
    pseudo_classes = widget.get_pseudo_class_state()
    assert pseudo_classes == PseudoClasses(enabled=True, focus=False, hover=False)


def test_get_pseudo_class_state_disabled():
    widget = Widget(disabled=True)
    pseudo_classes = widget.get_pseudo_class_state()
    assert pseudo_classes == PseudoClasses(enabled=False, focus=False, hover=False)


def test_get_pseudo_class_state_parent_disabled():
    child = Widget()
    _parent = Widget(disabled=True)
    child._attach(_parent)
    pseudo_classes = child.get_pseudo_class_state()
    assert pseudo_classes == PseudoClasses(enabled=False, focus=False, hover=False)


def test_get_pseudo_class_state_hover():
    widget = Widget()
    widget.mouse_hover = True
    pseudo_classes = widget.get_pseudo_class_state()
    assert pseudo_classes == PseudoClasses(enabled=True, focus=False, hover=True)


def test_get_pseudo_class_state_focus():
    widget = Widget()
    widget.has_focus = True
    pseudo_classes = widget.get_pseudo_class_state()
    assert pseudo_classes == PseudoClasses(enabled=True, focus=True, hover=False)


# Regression test for https://github.com/Textualize/textual/issues/1634
async def test_remove():
    class RemoveMeLabel(Label):
        async def on_mount(self) -> None:
            await self.run_action("app.remove_all")

    class Container(Widget):
        async def clear(self) -> None:
            await self.query("*").remove()

    class RemoveApp(App):
        def compose(self) -> ComposeResult:
            yield Container(RemoveMeLabel())

        async def action_remove_all(self) -> None:
            await self.query_one(Container).clear()
            self.exit(123)

    app = RemoveApp()
    async with app.run_test() as pilot:
        await pilot.press("r")
        await pilot.pause()
    assert app.return_value == 123


# Regression test for https://github.com/Textualize/textual/issues/2079
async def test_remove_unmounted():
    mounted = False

    class RemoveApp(App):
        def on_mount(self):
            nonlocal mounted
            label = Label()
            label.remove()
            mounted = True

    app = RemoveApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        assert mounted


def test_render_str() -> None:
    widget = Label()
    assert widget.render_str("foo") == Text("foo")
    assert widget.render_str("[b]foo") == Text.from_markup("[b]foo")
    # Text objects are passed unchanged
    text = Text("bar")
    assert widget.render_str(text) is text


async def test_compose_order() -> None:
    from textual.containers import Horizontal
    from textual.screen import Screen
    from textual.widgets import Select

    class MyScreen(Screen):
        def on_mount(self) -> None:
            self.query_one(Select).value = 1

        def compose(self) -> ComposeResult:
            yield Horizontal(
                Select(((str(n), n) for n in range(10)), id="select"),
                id="screen-horizontal",
            )

    class SelectBugApp(App[None]):
        async def on_mount(self):
            await self.push_screen(MyScreen(id="my-screen"))
            self.query_one(Select)

    app = SelectBugApp()
    messages: list[Message] = []

    async with app.run_test(message_hook=messages.append) as pilot:
        await pilot.pause()

    mounts = [
        message._sender.id
        for message in messages
        if isinstance(message, events.Mount) and message._sender.id is not None
    ]

    expected = [
        "_default",  # default  screen
        "label",  # A static in select
        "select",  # The select
        "screen-horizontal",  # The horizontal in MyScreen.compose
        "my-screen",  # THe screen mounted in the app
    ]

    assert mounts == expected


def test_children_must_be_widgets():
    with pytest.raises(TypeError):
        Widget(1, 2, 3)


def test_orphan_widget_has_no_siblings():
    assert Widget().siblings == []


def test__allow_scroll_default():
    assert not Widget()._allow_scroll


async def test__allow_scroll():
    from textual.containers import ScrollableContainer

    class AllowScrollApp(App):
        CSS = "ScrollableContainer { width: 3; height: 3; }"

        def compose(self):
            with ScrollableContainer():
                yield Label("This is\n\n\n\n\nlarge text.")

    app = AllowScrollApp()
    async with app.run_test():
        assert app.query_one(ScrollableContainer)._allow_scroll


async def test_offset_getter_setter():
    class OffsetApp(App):
        def compose(self):
            yield Label("hello")

    app = OffsetApp()
    async with app.run_test():
        label = app.query_one(Label)
        assert label.offset == Offset(0, 0)
        label.offset = (7, 3)
        assert label.offset == Offset(7, 3)


def test_get_set_tooltip():
    widget = Widget()
    assert widget.tooltip is None
    widget.tooltip = "This is a tooltip."
    assert widget.tooltip == "This is a tooltip."


async def test_loading():
    """Test setting the loading reactive."""

    class LoadingApp(App):
        def compose(self) -> ComposeResult:
            yield Label("Hello, World")

    async with LoadingApp().run_test() as pilot:
        app = pilot.app
        label = app.query_one(Label)
        assert label.loading == False
        assert len(label.query(LoadingIndicator)) == 0

        label.loading = True
        await pilot.pause()
        assert len(label.query(LoadingIndicator)) == 1

        label.loading = True  # Setting to same value is a null-op
        await pilot.pause()
        assert len(label.query(LoadingIndicator)) == 1

        label.loading = False
        await pilot.pause()
        assert len(label.query(LoadingIndicator)) == 0

        label.loading = False  # Setting to same value is a null-op
        await pilot.pause()
        assert len(label.query(LoadingIndicator)) == 0


async def test_is_mounted_property():
    class TestWidgetIsMountedApp(App):
        pass

    async with TestWidgetIsMountedApp().run_test() as pilot:
        widget = Widget()
        assert widget.is_mounted is False
        await pilot.app.mount(widget)
        assert widget.is_mounted is True


async def test_mount_error_not_widget():
    class NotWidgetApp(App):
        def compose(self) -> ComposeResult:
            yield {}

    app = NotWidgetApp()
    with pytest.raises(MountError):
        async with app.run_test():
            pass


async def test_mount_error_bad_widget():
    class DaftWidget(Widget):
        def __init__(self):
            # intentionally missing super()
            pass

    class NotWidgetApp(App):
        def compose(self) -> ComposeResult:
            yield DaftWidget()

    app = NotWidgetApp()
    with pytest.raises(MountError):
        async with app.run_test():
            pass


async def test_render_returns_text():
    """Test that render processes console markup when returning a string."""

    # Regression test for https://github.com/Textualize/textual/issues/3918
    class SimpleWidget(Widget):
        def render(self) -> str:
            return "Hello [b]World[/b]!"

    widget = SimpleWidget()
    render_result = widget._render()
    assert isinstance(render_result, Text)
    assert render_result.plain == "Hello World!"


async def test_sort_children() -> None:
    """Test the sort_children method."""

    class SortApp(App):
        def compose(self) -> ComposeResult:
            with Container(id="container"):
                yield Label("three", id="l3")
                yield Label("one", id="l1")
                yield Label("four", id="l4")
                yield Label("two", id="l2")

    app = SortApp()
    async with app.run_test():
        container = app.query_one("#container", Container)
        assert [label.id for label in container.query(Label)] == [
            "l3",
            "l1",
            "l4",
            "l2",
        ]
        container.sort_children(key=attrgetter("id"))
        assert [label.id for label in container.query(Label)] == [
            "l1",
            "l2",
            "l3",
            "l4",
        ]
        container.sort_children(key=attrgetter("id"), reverse=True)
        assert [label.id for label in container.query(Label)] == [
            "l4",
            "l3",
            "l2",
            "l1",
        ]


async def test_sort_children_no_key() -> None:
    """Test sorting with no key."""

    class SortApp(App):
        def compose(self) -> ComposeResult:
            with Container(id="container"):
                yield Label("three", id="l3")
                yield Label("one", id="l1")
                yield Label("four", id="l4")
                yield Label("two", id="l2")

    app = SortApp()
    async with app.run_test():
        container = app.query_one("#container", Container)
        assert [label.id for label in container.query(Label)] == [
            "l3",
            "l1",
            "l4",
            "l2",
        ]
        # Without a key, the sort order is the order children were instantiated
        container.sort_children()
        assert [label.id for label in container.query(Label)] == [
            "l3",
            "l1",
            "l4",
            "l2",
        ]
        container.sort_children(reverse=True)
        assert [label.id for label in container.query(Label)] == [
            "l2",
            "l4",
            "l1",
            "l3",
        ]


def test_bad_widget_name_raised() -> None:
    """Ensure error is raised when bad class names are used for widgets."""

    with pytest.raises(BadWidgetName):

        class lowercaseWidget(Widget):
            pass



from textual import events
from textual.app import App
from textual.widgets import Input


async def test_paste_app():
    paste_events = []

    class PasteApp(App):
        def on_paste(self, event):
            paste_events.append(event)

    app = PasteApp()
    async with app.run_test() as pilot:
        app.post_message(events.Paste(text="Hello"))
        await pilot.pause(0)

    assert len(paste_events) == 1
    assert paste_events[0].text == "Hello"


async def test_empty_paste():
    """Regression test for https://github.com/Textualize/textual/issues/2563."""

    paste_events = []

    class MyInput(Input):
        def on_paste(self, event):
            super()._on_paste(event)
            paste_events.append(event)

    class PasteApp(App):
        def compose(self):
            yield MyInput()

        def key_p(self):
            self.query_one(MyInput).post_message(events.Paste(""))

    app = PasteApp()
    async with app.run_test() as pilot:
        app.set_focus(None)
        await pilot.press("p")
        assert app.query_one(MyInput).value == ""
        assert len(paste_events) == 1
        assert paste_events[0].text == ""



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Iterable, NamedTuple

import rich.repr
from rich.console import RenderableType
from rich.measure import Measurement
from rich.padding import Padding
from rich.rule import Rule
from rich.style import NULL_STYLE, Style

from textual import _widget_navigation, events
from textual._widget_navigation import Direction
from textual.binding import Binding, BindingType
from textual.cache import LRUCache
from textual.geometry import Region, Size
from textual.message import Message
from textual.reactive import reactive
from textual.scroll_view import ScrollView
from textual.strip import Strip

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias


class DuplicateID(Exception):
    """Raised if a duplicate ID is used when adding options to an option list."""


class OptionDoesNotExist(Exception):
    """Raised when a request has been made for an option that doesn't exist."""


class Separator:
    """Class used to add a separator to an [OptionList][textual.widgets.OptionList]."""


@rich.repr.auto
class Option:
    """Class that holds the details of an individual option."""

    def __init__(
        self, prompt: RenderableType, id: str | None = None, disabled: bool = False
    ) -> None:
        """Initialise the option.

        Args:
            prompt: The prompt for the option.
            id: The optional ID for the option.
            disabled: The initial enabled/disabled state. Enabled by default.
        """
        self.__prompt = prompt
        self.__id = id
        self.disabled = disabled

    @property
    def prompt(self) -> RenderableType:
        """The prompt for the option."""
        return self.__prompt

    def set_prompt(self, prompt: RenderableType) -> None:
        """Set the prompt for the option.

        Args:
            prompt: The new prompt for the option.
        """
        self.__prompt = prompt

    @property
    def id(self) -> str | None:
        """The optional ID for the option."""
        return self.__id

    def __rich_repr__(self) -> rich.repr.Result:
        yield "prompt", self.prompt
        yield "id", self.id, None
        yield "disabled", self.disabled, False

    def __rich__(self) -> RenderableType:
        return self.__prompt


class OptionLineSpan(NamedTuple):
    """Class that holds the line span information for an option.

    An [Option][textual.widgets.option_list.Option] can have a prompt that
    spans multiple lines. Also, there's no requirement that every option in
    an option list has the same span information. So this structure is used
    to track the line that an option starts on, and how many lines it
    contains.
    """

    first: int
    """The line position for the start of the option.."""
    line_count: int
    """The count of lines that make up the option."""


OptionListContent: TypeAlias = "Option | Separator"
"""The type of an item of content in the option list.

This type represents all of the types that will be found in the list of
content of the option list after it has been processed for addition.
"""

NewOptionListContent: TypeAlias = "OptionListContent | None | RenderableType"
"""The type of a new item of option list content to be added to an option list.

This type represents all of the types that will be accepted when adding new
content to the option list. This is a superset of [`OptionListContent`][textual.types.OptionListContent].
"""


class OptionList(ScrollView, can_focus=True):
    """A vertical option list with bounce-bar highlighting."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("down", "cursor_down", "Down", show=False),
        Binding("end", "last", "Last", show=False),
        Binding("enter", "select", "Select", show=False),
        Binding("home", "first", "First", show=False),
        Binding("pagedown", "page_down", "Page Down", show=False),
        Binding("pageup", "page_up", "Page Up", show=False),
        Binding("up", "cursor_up", "Up", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | down | Move the highlight down. |
    | end | Move the highlight to the last option. |
    | enter | Select the current option. |
    | home | Move the highlight to the first option. |
    | pagedown | Move the highlight down a page of options. |
    | pageup | Move the highlight up a page of options. |
    | up | Move the highlight up. |
    """

    DEFAULT_CSS = """
    OptionList {
        height: auto;
        max-height: 100%;
        background: $boost;
        color: $text;
        overflow-x: hidden;
        border: tall transparent;
        padding: 0 1;
    }

    OptionList:focus {
        border: tall $accent;

    }

    OptionList > .option-list--separator {
        color: $foreground 15%;
    }

    OptionList > .option-list--option-highlighted {
        color: $text;
        text-style: bold;
    }

    OptionList:focus > .option-list--option-highlighted {
        background: $accent;
    }

    OptionList > .option-list--option-disabled {
        color: $text-disabled;
    }

    OptionList > .option-list--option-hover {
        background: $boost;
    }

    OptionList > .option-list--option-hover-highlighted {
        background: $accent 60%;
        color: $text;
        text-style: bold;
    }

    OptionList:focus > .option-list--option-hover-highlighted {
        background: $accent;
        color: $text;
        text-style: bold;
    }
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "option-list--option",
        "option-list--option-disabled",
        "option-list--option-highlighted",
        "option-list--option-hover",
        "option-list--option-hover-highlighted",
        "option-list--separator",
    }
    """
    | Class | Description |
    | :- | :- |
    | `option-list--option-disabled` | Target disabled options. |
    | `option-list--option-highlighted` | Target the highlighted option. |
    | `option-list--option-hover` | Target an option that has the mouse over it. |
    | `option-list--option-hover-highlighted` | Target a highlighted option that has the mouse over it. |
    | `option-list--separator` | Target the separators. |
    """

    highlighted: reactive[int | None] = reactive["int | None"](None)
    """The index of the currently-highlighted option, or `None` if no option is highlighted."""

    class OptionMessage(Message):
        """Base class for all option messages."""

        def __init__(self, option_list: OptionList, index: int) -> None:
            """Initialise the option message.

            Args:
                option_list: The option list that owns the option.
                index: The index of the option that the message relates to.
            """
            super().__init__()
            self.option_list: OptionList = option_list
            """The option list that sent the message."""
            self.option: Option = option_list.get_option_at_index(index)
            """The highlighted option."""
            self.option_id: str | None = self.option.id
            """The ID of the option that the message relates to."""
            self.option_index: int = index
            """The index of the option that the message relates to."""

        @property
        def control(self) -> OptionList:
            """The option list that sent the message.

            This is an alias for [`OptionMessage.option_list`][textual.widgets.OptionList.OptionMessage.option_list]
            and is used by the [`on`][textual.on] decorator.
            """
            return self.option_list

        def __rich_repr__(self) -> rich.repr.Result:
            yield "option_list", self.option_list
            yield "option", self.option
            yield "option_id", self.option_id
            yield "option_index", self.option_index

    class OptionHighlighted(OptionMessage):
        """Message sent when an option is highlighted.

        Can be handled using `on_option_list_option_highlighted` in a subclass of
        `OptionList` or in a parent node in the DOM.
        """

    class OptionSelected(OptionMessage):
        """Message sent when an option is selected.

        Can be handled using `on_option_list_option_selected` in a subclass of
        `OptionList` or in a parent node in the DOM.
        """

    def __init__(
        self,
        *content: NewOptionListContent,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        wrap: bool = True,
        tooltip: RenderableType | None = None,
    ):
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

        self._wrap = wrap
        """Should we auto-wrap options?

        If `False` options wider than the list will be truncated.
        """

        self._contents: list[OptionListContent] = [
            self._make_content(item) for item in content
        ]
        """A list of the content of the option list.

        This is *every* item that makes up the content of the option list;
        this includes both the options *and* the separators (and any other
        decoration we could end up adding -- although I don't anticipate
        anything else at the moment; but padding around separators could be
        a thing, perhaps).
        """

        self._options: list[Option] = [
            content for content in self._contents if isinstance(content, Option)
        ]
        """A list of the options within the option list.

        This is a list of references to just the options alone, ignoring the
        separators and potentially any other line-oriented option list
        content that isn't an option.
        """

        self._option_ids: dict[str, int] = {
            option.id: index
            for index, option in enumerate(self._options)
            if option.id is not None
        }
        """A dictionary of option IDs and the option indexes they relate to."""

        self._content_render_cache: LRUCache[tuple[int, Style, int], list[Strip]]
        self._content_render_cache = LRUCache(256)

        self._lines: list[tuple[int, int]] | None = None
        self._spans: list[OptionLineSpan] | None = None

        self._mouse_hovering_over: int | None = None
        """Used to track what the mouse is hovering over."""

        if tooltip is not None:
            self.tooltip = tooltip

        if self._options:
            self.action_first()

    def _left_gutter_width(self) -> int:
        """Returns the size of any left gutter that should be taken into account.

        Returns:
            The width of the left gutter.
        """
        return 0

    def _on_mount(self):
        self._populate()

    def _refresh_lines(self) -> None:
        self._lines = None
        self._spans = None
        self._content_render_cache.clear()
        self.check_idle()

    def notify_style_update(self) -> None:
        self._content_render_cache.clear()

    def _on_resize(self):
        self._refresh_lines()

    def on_idle(self):
        if self._lines is None:
            self._populate()

    def _add_lines(
        self, new_content: list[OptionListContent], width: int, option_index=0
    ) -> None:
        """Add new lines.

        Args:
            new_content: New content to add.
            width: Width to render content.
            option_index: Starting option index.
        """
        assert self._lines is not None
        assert self._spans is not None
        style = NULL_STYLE

        for index, content in enumerate(new_content, len(self._lines)):
            if isinstance(content, Option):
                height = len(
                    self._render_option_content(
                        index, content, style, width - self._left_gutter_width()
                    )
                )

                self._spans.append(OptionLineSpan(len(self._lines), height))
                self._lines.extend(
                    (option_index, y_offset) for y_offset in range(height)
                )
                option_index += 1
            else:
                self._lines.append(OptionLineSpan(-1, 0))

        self.virtual_size = Size(width, len(self._lines))

    def _populate(self) -> None:
        """Populate the lines data-structure."""
        if self._lines is not None:
            return
        self._lines = []
        self._spans = []

        self._add_lines(
            self._contents,
            self.scrollable_content_region.width - self._left_gutter_width(),
        )
        self.refresh()

    def get_content_width(self, container: Size, viewport: Size) -> int:
        """Get maximum width of options."""
        console = self.app.console
        options = console.options
        return max(
            Measurement.get(console, options, option.prompt).maximum
            for option in self._options
        )

    def get_content_height(self, container: Size, viewport: Size, width: int) -> int:
        # Get the content height without requiring a refresh
        # TODO: Internal data structure could be simplified
        style = self.rich_style
        _render_option_content = self._render_option_content
        heights = [
            len(_render_option_content(index, option, style, width))
            for index, option in enumerate(self._options)
        ]
        separator_count = sum(
            1 for content in self._contents if isinstance(content, Separator)
        )
        return sum(heights) + separator_count

    def _on_mouse_move(self, event: events.MouseMove) -> None:
        """React to the mouse moving.

        Args:
            event: The mouse movement event.
        """
        self._mouse_hovering_over = event.style.meta.get("option")

    def _on_leave(self, _: events.Leave) -> None:
        """React to the mouse leaving the widget."""
        self._mouse_hovering_over = None

    async def _on_click(self, event: events.Click) -> None:
        """React to the mouse being clicked on an item.

        Args:
            event: The click event.
        """
        clicked_option: int | None = event.style.meta.get("option")
        if (
            clicked_option is not None
            and clicked_option >= 0
            and not self._options[clicked_option].disabled
        ):
            self.highlighted = clicked_option
            self.action_select()

    def _make_content(self, content: NewOptionListContent) -> OptionListContent:
        """Convert a single item of content for the list into a content type.

        Args:
            content: The content to turn into a full option list type.

        Returns:
            The content, usable in the option list.
        """
        if isinstance(content, (Option, Separator)):
            return content
        if content is None:
            return Separator()
        return Option(content)

    def _render_option_content(
        self, option_index: int, renderable: RenderableType, style: Style, width: int
    ) -> list[Strip]:
        """Render content for option and style.

        Args:
            option_index: Option index to render.
            renderable: The Option renderable.
            style: The Rich style to render with.
            width: The width of the renderable.

        Returns:
            A list of strips.
        """
        cache_key = (option_index, style, width)
        if (strips := self._content_render_cache.get(cache_key, None)) is not None:
            return strips

        padding = self.get_component_styles("option-list--option").padding
        console = self.app.console
        options = console.options.update_width(width)
        if not self._wrap:
            options = options.update(no_wrap=True, overflow="ellipsis")
        if padding:
            renderable = Padding(renderable, padding)
        lines = self.app.console.render_lines(renderable, options, style=style)

        style_meta = Style.from_meta({"option": option_index})
        strips = [Strip(line, width).apply_style(style_meta) for line in lines]

        self._content_render_cache[cache_key] = strips
        return strips

    def _duplicate_id_check(self, candidate_items: list[OptionListContent]) -> None:
        """Check the items to be added for any duplicates.

        Args:
            candidate_items: The items that are going be added.

        Raises:
            DuplicateID: If there is an attempt to use a duplicate ID.
        """
        # We're only interested in options, and only those that have IDs.
        new_options = [
            item
            for item in candidate_items
            if isinstance(item, Option) and item.id is not None
        ]
        # Get the set of new IDs that we're being given.
        new_option_ids = {option.id for option in new_options}
        # Now check for duplicates, both internally amongst the new items
        # incoming, and also against all the current known IDs.
        if len(new_options) != len(new_option_ids) or not new_option_ids.isdisjoint(
            self._option_ids
        ):
            raise DuplicateID("Attempt made to add options with duplicate IDs.")

    def add_options(self, items: Iterable[NewOptionListContent]) -> Self:
        """Add new options to the end of the option list.

        Args:
            items: The new items to add.

        Returns:
            The `OptionList` instance.

        Raises:
            DuplicateID: If there is an attempt to use a duplicate ID.

        Note:
            All options are checked for duplicate IDs *before* any option is
            added. A duplicate ID will cause none of the passed items to be
            added to the option list.
        """
        # Only work if we have items to add; but don't make a fuss out of
        # zero items to add, just carry on like nothing happened.
        if self._lines is None:
            self._lines = []
        if self._spans is None:
            self._spans = []
        new_items = list(items)
        if new_items:
            option_index = len(self._options)
            # Turn any incoming values into valid content for the list.
            content = [self._make_content(item) for item in new_items]
            self._duplicate_id_check(content)
            self._contents.extend(content)
            # Pull out the content that is genuine options, create any new
            # ID mappings required, then add the new options to the option
            # list.
            new_options = [item for item in content if isinstance(item, Option)]
            for new_option_index, new_option in enumerate(
                new_options, start=len(self._options)
            ):
                if new_option.id:
                    self._option_ids[new_option.id] = new_option_index
            self._options.extend(new_options)

            self._add_lines(
                content,
                self.scrollable_content_region.width - self._left_gutter_width(),
                option_index=option_index,
            )
            self.refresh()
        return self

    def add_option(self, item: NewOptionListContent = None) -> Self:
        """Add a new option to the end of the option list.

        Args:
            item: The new item to add.

        Returns:
            The `OptionList` instance.

        Raises:
            DuplicateID: If there is an attempt to use a duplicate ID.
        """
        return self.add_options([item])

    def _remove_option(self, index: int) -> None:
        """Remove an option from the option list.

        Args:
            index: The index of the item to remove.

        Raises:
            IndexError: If there is no option of the given index.
        """
        option = self._options[index]
        del self._options[index]
        del self._contents[self._contents.index(option)]
        # Decrement index of options after the one we just removed.
        self._option_ids = {
            option_id: option_index - 1 if option_index > index else option_index
            for option_id, option_index in self._option_ids.items()
            if option_index != index
        }
        self._refresh_lines()
        # Force a re-validation of the highlight.
        self.highlighted = self.highlighted
        self._mouse_hovering_over = None

    def remove_option(self, option_id: str) -> Self:
        """Remove the option with the given ID.

        Args:
            option_id: The ID of the option to remove.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        self._remove_option(self.get_option_index(option_id))
        return self

    def remove_option_at_index(self, index: int) -> Self:
        """Remove the option at the given index.

        Args:
            index: The index of the option to remove.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        try:
            self._remove_option(index)
        except IndexError:
            raise OptionDoesNotExist(
                f"There is no option with an index of {index!r}"
            ) from None
        return self

    def _replace_option_prompt(self, index: int, prompt: RenderableType) -> None:
        """Replace the prompt of an option in the list.

        Args:
            index: The index of the option to replace the prompt of.
            prompt: The new prompt for the option.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        self.get_option_at_index(index).set_prompt(prompt)
        self._refresh_lines()

    def replace_option_prompt(self, option_id: str, prompt: RenderableType) -> Self:
        """Replace the prompt of the option with the given ID.

        Args:
            option_id: The ID of the option to replace the prompt of.
            prompt: The new prompt for the option.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        self._replace_option_prompt(self.get_option_index(option_id), prompt)
        return self

    def replace_option_prompt_at_index(
        self, index: int, prompt: RenderableType
    ) -> Self:
        """Replace the prompt of the option at the given index.

        Args:
            index: The index of the option to replace the prompt of.
            prompt: The new prompt for the option.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        self._replace_option_prompt(index, prompt)
        return self

    def clear_options(self) -> Self:
        """Clear the content of the option list.

        Returns:
            The `OptionList` instance.
        """
        self._contents.clear()
        self._options.clear()
        self._option_ids.clear()
        self.highlighted = None
        self._mouse_hovering_over = None
        self._refresh_lines()
        return self

    def _set_option_disabled(self, index: int, disabled: bool) -> Self:
        """Set the disabled state of an option in the list.

        Args:
            index: The index of the option to set the disabled state of.
            disabled: The disabled state to set.

        Returns:
            The `OptionList` instance.
        """
        self._options[index].disabled = disabled
        if index == self.highlighted:
            self.highlighted = _widget_navigation.find_next_enabled(
                self._options, anchor=index, direction=1
            )
        # TODO: Refresh only if the affected option is visible.
        self.refresh()
        return self

    def enable_option_at_index(self, index: int) -> Self:
        """Enable the option at the given index.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        try:
            return self._set_option_disabled(index, False)
        except IndexError:
            raise OptionDoesNotExist(
                f"There is no option with an index of {index}"
            ) from None

    def disable_option_at_index(self, index: int) -> Self:
        """Disable the option at the given index.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        try:
            return self._set_option_disabled(index, True)
        except IndexError:
            raise OptionDoesNotExist(
                f"There is no option with an index of {index}"
            ) from None

    def enable_option(self, option_id: str) -> Self:
        """Enable the option with the given ID.

        Args:
            option_id: The ID of the option to enable.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        return self.enable_option_at_index(self.get_option_index(option_id))

    def disable_option(self, option_id: str) -> Self:
        """Disable the option with the given ID.

        Args:
            option_id: The ID of the option to disable.

        Returns:
            The `OptionList` instance.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        return self.disable_option_at_index(self.get_option_index(option_id))

    @property
    def option_count(self) -> int:
        """The count of options."""
        return len(self._options)

    def get_option_at_index(self, index: int) -> Option:
        """Get the option at the given index.

        Args:
            index: The index of the option to get.

        Returns:
            The option at that index.

        Raises:
            OptionDoesNotExist: If there is no option with the given index.
        """
        try:
            return self._options[index]
        except IndexError:
            raise OptionDoesNotExist(
                f"There is no option with an index of {index}"
            ) from None

    def get_option(self, option_id: str) -> Option:
        """Get the option with the given ID.

        Args:
            option_id: The ID of the option to get.

        Returns:
            The option with the ID.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        return self.get_option_at_index(self.get_option_index(option_id))

    def get_option_index(self, option_id: str) -> int:
        """Get the index of the option with the given ID.

        Args:
            option_id: The ID of the option to get the index of.

        Returns:
            The index of the item with the given ID.

        Raises:
            OptionDoesNotExist: If no option has the given ID.
        """
        try:
            return self._option_ids[option_id]
        except KeyError:
            raise OptionDoesNotExist(
                f"There is no option with an ID of '{option_id}'"
            ) from None

    def render_line(self, y: int) -> Strip:
        self._populate()
        assert self._lines is not None

        _scroll_x, scroll_y = self.scroll_offset
        line_number = scroll_y + y

        try:
            option_index, y_offset = self._lines[line_number]
        except IndexError:
            return Strip([])

        renderable = (
            Rule(style=self.get_component_rich_style("option-list--separator"))
            if option_index == -1
            else self._options[option_index]
        )

        mouse_over = self._mouse_hovering_over == option_index

        component_class: str | None = None

        if option_index == -1:
            component_class = "option-list--separator"
        else:
            try:
                option = self._options[option_index]
            except IndexError:
                pass
            else:
                if option.disabled:
                    component_class = "option-list--option-disabled"
                elif self.highlighted == option_index:
                    component_class = "option-list--option-highlighted"
                elif mouse_over:
                    component_class = "option-list--option-hover"

        style = (
            self.get_component_rich_style(component_class)
            if component_class
            else self.rich_style
        )

        strips = self._render_option_content(
            option_index,
            renderable,
            style,
            self.scrollable_content_region.width - self._left_gutter_width(),
        )
        try:
            strip = strips[y_offset]
        except IndexError:
            return Strip([])
        return strip

    def scroll_to_highlight(self, top: bool = False) -> None:
        """Ensure that the highlighted option is in view.

        Args:
            top: Scroll highlight to top of the list.
        """
        highlighted = self.highlighted
        if highlighted is None or self._spans is None:
            return

        try:
            y, height = self._spans[highlighted]
        except IndexError:
            # Index error means we're being asked to scroll to a highlight
            # before all the tracking information has been worked out.
            # That's fine; let's just NoP that.
            return
        self.scroll_to_region(
            Region(0, y, self.scrollable_content_region.width, height),
            force=True,
            animate=False,
            top=top,
        )

    def validate_highlighted(self, highlighted: int | None) -> int | None:
        """Validate the `highlighted` property value on access."""
        if highlighted is None or not self._options:
            return None
        elif highlighted < 0:
            return 0
        elif highlighted >= len(self._options):
            return len(self._options) - 1

        return highlighted

    def watch_highlighted(self, highlighted: int | None) -> None:
        """React to the highlighted option having changed."""
        if highlighted is not None and not self._options[highlighted].disabled:
            self.scroll_to_highlight()
            self.post_message(self.OptionHighlighted(self, highlighted))

    def action_cursor_up(self) -> None:
        """Move the highlight up to the previous enabled option."""
        self.highlighted = _widget_navigation.find_next_enabled(
            self._options,
            anchor=self.highlighted,
            direction=-1,
        )

    def action_cursor_down(self) -> None:
        """Move the highlight down to the next enabled option."""
        self.highlighted = _widget_navigation.find_next_enabled(
            self._options,
            anchor=self.highlighted,
            direction=1,
        )

    def action_first(self) -> None:
        """Move the highlight to the first enabled option."""
        self.highlighted = _widget_navigation.find_first_enabled(self._options)

    def action_last(self) -> None:
        """Move the highlight to the last enabled option."""
        self.highlighted = _widget_navigation.find_last_enabled(self._options)

    def _page(self, direction: Direction) -> None:
        """Move the highlight roughly by one page in the given direction.

        The highlight will tentatively move by exactly one page.
        If this would result in highlighting a disabled option, instead we look for
        an enabled option "further down" the list of options.
        If there are no such enabled options, we fallback to the "last" enabled option.
        (The meaning of "further down" and "last" depend on the direction specified.)

        Args:
            direction: The direction to head, -1 for up and 1 for down.
        """

        # If we find ourselves in a position where we don't know where we're
        # going, we need a fallback location. Where we go will depend on the
        # direction.
        self._populate()
        assert self._spans is not None
        assert self._lines is not None

        fallback = self.action_first if direction == -1 else self.action_last

        highlighted = self.highlighted
        if highlighted is None:
            # There is no highlight yet so let's go to the default position.
            fallback()
        else:
            # We want to page roughly by lines, but we're dealing with
            # options that can be a varying number of lines in height. So
            # let's start with the target line alone.
            target_line = max(
                0,
                self._spans[highlighted].first
                + (direction * self.scrollable_content_region.height),
            )
            try:
                # Now that we've got a target line, let's figure out the
                # index of the target option.
                target_option: int | None = self._lines[target_line][0]
            except IndexError:
                # An index error suggests we've gone out of bounds, let's
                # settle on whatever the call thinks is a good place to wrap
                # to.
                fallback()
            else:
                # Looks like we've figured where we'd like to jump to, we
                # just need to make sure we jump to an option that's enabled.
                if target_option is not None:
                    target_option = _widget_navigation.find_next_enabled_no_wrap(
                        candidates=self._options,
                        anchor=target_option,
                        direction=direction,
                        with_anchor=True,
                    )
                    # If we couldn't find an enabled option that's at least one page
                    # away from the current one, we instead move less than one page
                    # to the last enabled option in the correct direction.
                    if target_option is None:
                        fallback()
                    else:
                        self.highlighted = target_option

    def action_page_up(self) -> None:
        """Move the highlight up roughly by one page."""
        self._page(-1)

    def action_page_down(self) -> None:
        """Move the highlight down roughly by one page."""
        self._page(1)

    def action_select(self) -> None:
        """Select the currently-highlighted option.

        If no option is selected, then nothing happens. If an option is
        selected, a [OptionList.OptionSelected][textual.widgets.OptionList.OptionSelected]
        message will be posted.
        """
        highlighted = self.highlighted
        if highlighted is not None and not self._options[highlighted].disabled:
            self.post_message(self.OptionSelected(self, highlighted))


if __name__ == "__main__":
    from textual.app import App, ComposeResult

    class OptionApp(App):
        def compose(self) -> ComposeResult:
            yield OptionList("Foo", "Bar", "Baz")

    app = OptionApp()
    app.run()



from time import perf_counter

from textual.app import App, ComposeResult
from textual.reactive import var
from textual.widgets import Static


class AnimApp(App):
    CSS = """
    #foo {
        height: 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield Static("foo", id="foo")


async def test_animate_height() -> None:
    """Test animating styles.height works."""

    # Styles.height is a scalar, which makes it more complicated to animate

    app = AnimApp()

    async with app.run_test() as pilot:
        static = app.query_one(Static)
        assert static.size.height == 1
        assert static.styles.height.value == 1
        static.styles.animate("height", 100, duration=0.5, easing="linear")
        start = perf_counter()

        # Wait for the animation to finished
        await pilot.wait_for_animation()
        elapsed = perf_counter() - start
        # Check that the full time has elapsed
        assert elapsed >= 0.5
        # Check the height reached the maximum
        assert static.styles.height.value == 100


async def test_scheduling_animation() -> None:
    """Test that scheduling an animation works."""

    app = AnimApp()
    delay = 0.1

    async with app.run_test() as pilot:
        styles = app.query_one(Static).styles
        styles.background = "black"

        styles.animate("background", "white", delay=delay, duration=0)

        # Still black immediately after call, animation hasn't started yet due to `delay`
        assert styles.background.rgb == (0, 0, 0)

        await pilot.wait_for_scheduled_animations()
        assert styles.background.rgb == (255, 255, 255)


async def test_wait_for_current_animations() -> None:
    """Test that we can wait only for the current animations taking place."""

    app = AnimApp()

    delay = 10

    async with app.run_test() as pilot:
        styles = app.query_one(Static).styles
        styles.animate("height", 100, duration=0.1)
        start = perf_counter()
        styles.animate("height", 200, duration=0.1, delay=delay)

        # Wait for the first animation to finish
        await pilot.wait_for_animation()
        elapsed = perf_counter() - start
        assert elapsed < (delay / 2)


async def test_wait_for_current_and_scheduled_animations() -> None:
    """Test that we can wait for current and scheduled animations."""

    app = AnimApp()

    async with app.run_test() as pilot:
        styles = app.query_one(Static).styles

        start = perf_counter()
        styles.animate("height", 50, duration=0.01)
        styles.animate("background", "black", duration=0.01, delay=0.05)

        await pilot.wait_for_scheduled_animations()
        elapsed = perf_counter() - start
        assert elapsed >= 0.06
        assert styles.background.rgb == (0, 0, 0)


async def test_reverse_animations() -> None:
    """Test that you can create reverse animations.

    Regression test for #1372 https://github.com/Textualize/textual/issues/1372
    """

    app = AnimApp()

    async with app.run_test() as pilot:
        static = app.query_one(Static)
        styles = static.styles

        # Starting point.
        styles.background = "black"
        assert styles.background.rgb == (0, 0, 0)

        # First, make sure we can go from black to white and back, step by step.
        styles.animate("background", "white", duration=0.01)
        await pilot.wait_for_animation()
        assert styles.background.rgb == (255, 255, 255)

        styles.animate("background", "black", duration=0.01)
        await pilot.wait_for_animation()
        assert styles.background.rgb == (0, 0, 0)

        # Now, the actual test is to make sure we go back to black if creating both at once.
        styles.animate("background", "white", duration=0.01)
        styles.animate("background", "black", duration=0.01)
        await pilot.wait_for_animation()
        assert styles.background.rgb == (0, 0, 0)


async def test_schedule_reverse_animations() -> None:
    """Test that you can schedule reverse animations.

    Regression test for #1372 https://github.com/Textualize/textual/issues/1372
    """

    app = AnimApp()

    async with app.run_test() as pilot:
        static = app.query_one(Static)
        styles = static.styles

        # Starting point.
        styles.background = "black"
        assert styles.background.rgb == (0, 0, 0)

        # First, make sure we can go from black to white and back, step by step.
        styles.animate("background", "white", delay=0.01, duration=0.01)
        await pilot.wait_for_scheduled_animations()
        assert styles.background.rgb == (255, 255, 255)

        styles.animate("background", "black", delay=0.01, duration=0.01)
        await pilot.wait_for_scheduled_animations()
        assert styles.background.rgb == (0, 0, 0)

        # Now, the actual test is to make sure we go back to black if scheduling both at once.
        styles.animate("background", "white", delay=0.025, duration=0.05)
        # While the black -> white animation runs, start the white -> black animation.
        styles.animate("background", "black", delay=0.05, duration=0.01)
        await pilot.wait_for_scheduled_animations()
        assert styles.background.rgb == (0, 0, 0)


class CancelAnimWidget(Static):
    counter: var[float] = var(23)


class CancelAnimApp(App[None]):
    counter: var[float] = var(23)

    def compose(self) -> ComposeResult:
        yield CancelAnimWidget()


async def test_cancel_app_animation() -> None:
    """It should be possible to cancel a running app animation."""

    async with CancelAnimApp().run_test() as pilot:
        pilot.app.animate("counter", value=0, final_value=1000, duration=60)
        await pilot.pause()
        assert pilot.app.animator.is_being_animated(pilot.app, "counter")
        await pilot.app.stop_animation("counter")
        assert not pilot.app.animator.is_being_animated(pilot.app, "counter")


async def test_cancel_app_non_animation() -> None:
    """It should be possible to attempt to cancel a non-running app animation."""

    async with CancelAnimApp().run_test() as pilot:
        assert not pilot.app.animator.is_being_animated(pilot.app, "counter")
        await pilot.app.stop_animation("counter")
        assert not pilot.app.animator.is_being_animated(pilot.app, "counter")


async def test_cancel_widget_animation() -> None:
    """It should be possible to cancel a running widget animation."""

    async with CancelAnimApp().run_test() as pilot:
        widget = pilot.app.query_one(CancelAnimWidget)
        widget.animate("counter", value=0, final_value=1000, duration=60)
        await pilot.pause()
        assert pilot.app.animator.is_being_animated(widget, "counter")
        await widget.stop_animation("counter")
        assert not pilot.app.animator.is_being_animated(widget, "counter")


async def test_cancel_widget_non_animation() -> None:
    """It should be possible to attempt to cancel a non-running widget animation."""

    async with CancelAnimApp().run_test() as pilot:
        widget = pilot.app.query_one(CancelAnimWidget)
        assert not pilot.app.animator.is_being_animated(widget, "counter")
        await widget.stop_animation("counter")
        assert not pilot.app.animator.is_being_animated(widget, "counter")



from textual.app import App, ComposeResult
from textual.color import Color, ColorParseError
from textual.containers import Grid
from textual.reactive import reactive
from textual.widgets import Input, Static


class WatchApp(App):
    CSS_PATH = "watch01.tcss"

    color = reactive(Color.parse("transparent"))  # (1)!

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Enter a color")
        yield Grid(Static(id="old"), Static(id="new"), id="colors")

    def watch_color(self, old_color: Color, new_color: Color) -> None:  # (2)!
        self.query_one("#old").styles.background = old_color
        self.query_one("#new").styles.background = new_color

    def on_input_submitted(self, event: Input.Submitted) -> None:
        try:
            input_color = Color.parse(event.value)
        except ColorParseError:
            pass
        else:
            self.query_one(Input).value = ""
            self.color = input_color  # (3)!


if __name__ == "__main__":
    app = WatchApp()
    app.run()



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import DataTable, Footer

MOVIES = [
    "Severance",
    "Foundation",
    "Dark",
    "The Boys",
    "The Last of Us",
    "Lost in Space",
    "Altered Carbon",
]


class AddColumn(App):
    BINDINGS = [
        Binding(key="c", action="add_column", description="Add Column"),
    ]

    def compose(self) -> ComposeResult:
        table = DataTable()
        table.add_column("Movies")
        row_keys = table.add_rows([(movie,) for movie in MOVIES])

        column_key = table.add_column("No Default")
        table.add_column("With Default", default="ABC")
        table.add_column("Long Default", default="01234567890123456789")

        # Ensure we can update a cell
        table.update_cell(row_keys[2], column_key, "Hello!")

        yield table


app = AddColumn()
if __name__ == '__main__':
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import LoadingIndicator


class LoadingApp(App):
    def compose(self) -> ComposeResult:
        yield LoadingIndicator()


if __name__ == "__main__":
    app = LoadingApp()
    app.run()



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Static

ERROR_TEXT = """
An error has occurred. To continue:

Press Enter to return to Windows, or

Press CTRL+ALT+DEL to restart your computer. If you do this,
you will lose any unsaved information in all open applications.

Error: 0E : 016F : BFF9B3D4
"""


class BSOD(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    def compose(self) -> ComposeResult:
        yield Static(" Windows ", id="title")
        yield Static(ERROR_TEXT)
        yield Static("Press any key to continue [blink]_[/]", id="any-key")


class BSODApp(App):
    CSS_PATH = "screen02.tcss"
    BINDINGS = [("b", "push_screen('bsod')", "BSOD")]

    def on_mount(self) -> None:
        self.install_screen(BSOD(), name="bsod")


if __name__ == "__main__":
    app = BSODApp()
    app.run()



from textual.app import App
from textual.containers import Container, Horizontal, ScrollableContainer, Vertical
from textual.screen import Screen
from textual.widgets import DataTable, Header, Label


class LabeledBox(Container):
    DEFAULT_CSS = """
    LabeledBox {
        layers: base_ top_;
        width: 100%;
        height: 100%;
    }

    LabeledBox > Container {
        layer: base_;
        border: round $primary;
        width: 100%;
        height: 100%;
        layout: vertical;
    }

    LabeledBox > Label {
        layer: top_;
        offset-x: 2;
    }
    """

    def __init__(self, title, *args, **kwargs):
        self.__label = Label(title)

        super().__init__(self.__label, Container(*args, **kwargs))

    @property
    def label(self):
        return self.__label


class StatusTable(DataTable):
    def __init__(self):
        super().__init__()

        self.cursor_type = "row"
        self.show_cursor = False
        self.add_column("Foo")
        self.add_column("Bar")
        self.add_column("Baz")

        for _ in range(50):
            self.add_row(
                "ABCDEFGH",
                "0123456789",
                "IJKLMNOPQRSTUVWXYZ",
            )


class Status(LabeledBox):
    DEFAULT_CSS = """
    Status {
        width: auto;
    }

    Status Container {
        width: auto;
    }

    Status StatusTable {
        width: auto;
        height: 100%;
        margin-top: 1;
        scrollbar-gutter: stable;
        overflow-x: hidden;
    }
    """

    def __init__(self, name: str):
        self.__name = name
        self.__table = StatusTable()

        super().__init__(f" {self.__name} ", self.__table)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def table(self) -> StatusTable:
        return self.__table


class Rendering(LabeledBox):
    DEFAULT_CSS = """
    #issue-info {
        height: auto;
        border-bottom: dashed #632CA6;
    }

    #statuses-box {
        height: 1fr;
        width: auto;
    }
    """

    def __init__(self):
        self.__info = Label("test")

        super().__init__(
            "",
            ScrollableContainer(
                Horizontal(self.__info, id="issue-info"),
                Horizontal(*[Status(str(i)) for i in range(4)], id="statuses-box"),
                id="issues-box",
            ),
        )

    @property
    def info(self) -> Label:
        return self.__info


class Sidebar(LabeledBox):
    DEFAULT_CSS = """
    #sidebar-status {
        height: auto;
        border-bottom: dashed #632CA6;
    }

    #sidebar-options {
        height: 1fr;
    }
    """

    def __init__(self):
        self.__status = Label("ok")
        self.__options = Vertical()

        super().__init__(
            "",
            Container(self.__status, id="sidebar-status"),
            Container(self.__options, id="sidebar-options"),
        )

    @property
    def status(self) -> Label:
        return self.__status

    @property
    def options(self) -> Vertical:
        return self.__options


class MyScreen(Screen):
    DEFAULT_CSS = """
    #main-content {
        layout: grid;
        grid-size: 2;
        grid-columns: 1fr 5fr;
        grid-rows: 1fr;
    }

    #main-content-sidebar {
        height: 100%;
    }

    #main-content-rendering {
        height: 100%;
    }
    """

    def compose(self):
        yield Header()
        yield Container(
            Container(Sidebar(), id="main-content-sidebar"),
            Container(Rendering(), id="main-content-rendering"),
            id="main-content",
        )


class MyApp(App):
    async def on_mount(self):
        self.install_screen(MyScreen(), "myscreen")
        await self.push_screen("myscreen")


if __name__ == "__main__":
    app = MyApp()
    app.run()



import pytest

from textual.app import App, ComposeResult
from textual.widgets import TextArea
from textual.widgets.text_area import (
    Document,
    LanguageDoesNotExist,
    SyntaxAwareDocument,
)


class TextAreaApp(App):
    def compose(self) -> ComposeResult:
        yield TextArea("print('hello')", language="python")


async def test_setting_builtin_language_via_constructor():
    class MyTextAreaApp(App):
        def compose(self) -> ComposeResult:
            yield TextArea("print('hello')", language="python")

    app = MyTextAreaApp()

    async with app.run_test():
        text_area = app.query_one(TextArea)
        assert text_area.language == "python"

        text_area.language = "markdown"
        assert text_area.language == "markdown"


async def test_setting_builtin_language_via_attribute():
    class MyTextAreaApp(App):
        def compose(self) -> ComposeResult:
            text_area = TextArea("print('hello')")
            text_area.language = "python"
            yield text_area

    app = MyTextAreaApp()

    async with app.run_test():
        text_area = app.query_one(TextArea)
        assert text_area.language == "python"

        text_area.language = "markdown"
        assert text_area.language == "markdown"


async def test_setting_language_to_none():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.language = None
        assert text_area.language is None


async def test_setting_unknown_language():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        with pytest.raises(LanguageDoesNotExist):
            text_area.language = "this-language-doesnt-exist"


@pytest.mark.syntax
async def test_register_language():
    app = TextAreaApp()

    async with app.run_test():
        text_area = app.query_one(TextArea)

        from tree_sitter_languages import get_language

        language = get_language("elm")

        # ...and register it with no highlights
        text_area.register_language(language, "")

        # Ensure that registered language is now available.
        assert "elm" in text_area.available_languages

        # Switch to the newly registered language
        text_area.language = "elm"

        assert text_area.language == "elm"


@pytest.mark.syntax
async def test_register_language_existing_language():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        # Before registering the language, we have highlights as expected.
        assert len(text_area._highlights) > 0

        # Overwriting the highlight query for Python...
        text_area.register_language("python", "")

        # We've overridden the highlight query with a blank one, so there are no highlights.
        assert text_area._highlights == {}


@pytest.mark.syntax
async def test_language_binary_missing(monkeypatch: pytest.MonkeyPatch):
    # mock a failed installation of tree-sitter-language binaries by
    # raising an OSError from get_language
    def raise_oserror(_):
        raise OSError(
            "/path/to/tree_sitter_languages/languages.so: "
            "cannot open shared object file: No such file or directory"
        )

    monkeypatch.setattr(
        "textual.document._syntax_aware_document.get_language", raise_oserror
    )

    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)  # does not crash
        assert text_area.language == "python"
        # resulting document is not a SyntaxAwareDocument and does not
        # support highlights
        assert isinstance(text_area.document, Document)
        assert not isinstance(text_area.document, SyntaxAwareDocument)
        assert text_area._highlights == {}



from textual.app import App
from textual.widgets import Label, Static
from rich.panel import Panel


class LabelWrap(App):
    CSS = """Screen {
              align: center middle;
          }

          #l_data {
              border: blank;
              background: lightgray;
          }

          #s_data {
              border: blank;
              background: lightgreen;
          }

          #p_data {
              border: blank;
              background: lightgray;
          }"""

    def __init__(self):
        super().__init__()

        self.data = (
            "Apple Banana Cherry Mango Fig Guava Pineapple:"
            "Dragon Unicorn Centaur Phoenix Chimera Castle"
        )

    def compose(self):
        yield Label(self.data, id="l_data")
        yield Static(self.data, id="s_data")
        yield Label(Panel(self.data), id="p_data")

    def on_mount(self):
        self.dark = False


if __name__ == "__main__":
    app = LabelWrap()
    app.run()



import pytest

from textual.screen import Screen
from textual.widget import Widget


@pytest.mark.parametrize(
    "layout,display,expected_in_displayed_children",
    [
        ("horizontal", "block", True),
        ("vertical", "block", True),
        ("horizontal", "none", False),
        ("vertical", "none", False),
    ],
)
def test_nodes_take_display_property_into_account_when_they_display_their_children(
    layout: str, display: str, expected_in_displayed_children: bool
):
    widget = Widget(name="widget that might not be visible 🥷 ")
    widget.styles.display = display

    screen = Screen()
    screen.styles.layout = layout
    screen._add_child(widget)

    displayed_children = screen.displayed_children
    assert isinstance(displayed_children, list)
    assert (widget in screen.displayed_children) is expected_in_displayed_children



import pytest

from textual._slug import TrackedSlugs, slug


@pytest.mark.xdist_group("group1")
@pytest.mark.parametrize(
    "text, expected",
    [
        ("test", "test"),
        ("Test", "test"),
        (" Test ", "test"),
        ("-test-", "-test-"),
        ("!test!", "test"),
        ("test!!test", "testtest"),
        ("test! !test", "test-test"),
        ("test test", "test-test"),
        ("test  test", "test--test"),
        ("test          test", "test----------test"),
        ("--test", "--test"),
        ("test--", "test--"),
        ("--test--test--", "--test--test--"),
        ("test!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~test", "test-_test"),
        ("tëst", "t%C3%ABst"),
        ("test🙂test", "testtest"),
        ("test🤷test", "testtest"),
        ("test🤷🏻‍♀️test", "testtest"),
    ],
)
def test_simple_slug(text: str, expected: str) -> None:
    """The simple slug function should produce the expected slug."""
    assert slug(text) == expected


@pytest.fixture(scope="module")
def tracker() -> TrackedSlugs:
    return TrackedSlugs()


@pytest.mark.xdist_group("group2")
@pytest.mark.parametrize(
    "text, expected",
    [
        ("test", "test"),
        ("test", "test-1"),
        ("test", "test-2"),
        ("-test-", "-test-"),
        ("-test-", "-test--1"),
        ("test!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~test", "test-_test"),
        ("test!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~test", "test-_test-1"),
        ("tëst", "t%C3%ABst"),
        ("tëst", "t%C3%ABst-1"),
        ("tëst", "t%C3%ABst-2"),
        ("test🙂test", "testtest"),
        ("test🤷test", "testtest-1"),
        ("test🤷🏻‍♀️test", "testtest-2"),
        ("test", "test-3"),
        ("test", "test-4"),
        (" test ", "test-5"),
    ],
)
def test_tracked_slugs(tracker: TrackedSlugs, text: str, expected: str) -> None:
    """The tracked slugging class should produce the expected slugs."""
    assert tracker.slug(text) == expected
